import numpy as np
from typing import Literal

from uplt.engine.plotly.scale import get_scale


def estimate_axis_range(figure,
                        axis: Literal['x', 'y', 'z'],
                        mode: Literal['min', 'max']) -> float:
    """
    Setting only min or only max is not implemented in plotly,
    so manual estimation of min/max is required:

        https://github.com/plotly/plotly.js/issues/400
        https://github.com/plotly/plotly.py/issues/3634
    """
    if figure.data is None or len(figure.data) == 0:
        raise RuntimeError('there is no any graph, use xlim/ylim after plotting or '
                           'specify both range_min and range_max')

    # estimate min/max from data
    data_minmax = []
    minmax_estimate = None
    for trace_data in figure.data:
        values = trace_data[axis]
        if np.issubdtype(values.dtype, np.number):
            # array of numbers
            minmax_estimate = { 'min': np.min, 'max': np.max }[mode]
        else:
            # array of str or other objects
            minmax_estimate = { 'min': lambda x: x[0], 'max': lambda x: x[-1] }[mode]

        data_minmax.append(minmax_estimate(values))

    assert minmax_estimate is not None
    minmax = minmax_estimate(data_minmax)

    # estimate min/max from range
    axis_name: str = {
        'x': 'xaxis',
        'y': 'yaxis'
    }[axis]

    axis_range = figure.layout[axis_name]['range']
    if axis_range is not None:
        if get_scale(figure, axis) == 'log':
            axis_range = 10**np.asarray(axis_range)
        minmax = minmax_estimate([ minmax, *axis_range ])

    return minmax



import numpy as np
from collections.abc import Sequence
from numpy.typing import ArrayLike
from plotly.graph_objs import Figure

import uplt.color as ucolor
from uplt.interface import LineStyle, MarkerStyle
from uplt.default import DEFAULT


def plot_line_marker(figure      : Figure,
                     color       : str | Sequence[str],
                     x           : ArrayLike,
                     y           : ArrayLike | None = None,
                     z           : ArrayLike | None = None,
                     name        : str | None = None,
                     line_style  : LineStyle | None = None,
                     line_width  : float = 2,
                     marker_style: MarkerStyle | None = None,
                     marker_size : float | None = None,
                     opacity     : float = 1.0,
                     legend_group: str | None = None,
                     legend_group_title: str | None = None,
                     **kwargs):
    """
    General plot: line, line+markers, markers(scatter).
    """
    x = np.atleast_1d(np.asarray(x))

    if y is None:
        y = x
        x = np.arange(len(y))
    else:
        y = np.asarray(y)

    y = np.atleast_1d(y)

    assert x.ndim == y.ndim == 1, 'the input must be 1d arrays'
    assert len(x) == len(y), 'the length of the input arrays must be the same'

    if z is not None:
        z = np.atleast_1d(np.asarray(z))
        assert z.ndim == 1, 'the input must be 1d arrays'
        assert len(x) == len(z), 'the length of the input arrays must be the same'

    if marker_size is None:
        # 1.33 - conversion gain between matplotlib and plotly marker size
        marker_size = DEFAULT.marker_size*1.33

    if name is None:
        name = ''
        show_legend = False
    else:
        show_legend = kwargs.pop('showlegend', True)

    if isinstance(color, str):
        color = ucolor.name_to_hex(color)
    else:
        # color specified for each point (x, y)
        color = [ ucolor.name_to_hex(c) for c in color ]

    from uplt.engine.plotly.mapping import LINE_STYLE_MAPPING, MARKER_STYLE_MAPPING
    line_style_str = LINE_STYLE_MAPPING[line_style]
    marker_style_str = MARKER_STYLE_MAPPING[marker_style]

    line = kwargs.pop('line', {})
    marker = kwargs.pop('marker', {})

    mode = 'lines' if marker_style_str is None else 'lines+markers'
    if line_style_str == ' ': # no lines = scatter mode
        mode = 'markers'
        line['dash'] = None
    else:
        line.setdefault('color', color)
        line.setdefault('width', line_width)
        line['dash'] = line_style_str

    marker.setdefault('color', color)
    marker.setdefault('line_color', color)
    marker.setdefault('line_width', line_width)
    marker['symbol'] = marker_style_str
    marker['size'] = marker_size

    hoverlabel = kwargs.pop('hoverlabel', {})
    hoverlabel.setdefault('namelength', -1)

    if z is None:
        figure.add_scatter(x=x, y=y,
                           name=name,
                           mode=mode,
                           line=line,
                           marker=marker,
                           opacity=opacity,
                           showlegend=show_legend,
                           hoverlabel=hoverlabel,
                           legendgroup=legend_group,
                           legendgrouptitle_text=legend_group_title,
                           **kwargs)
    else:
        figure.add_scatter3d(x=x, y=y, z=z,
                             name=name,
                             mode=mode,
                             line=line,
                             marker=marker,
                             opacity=opacity,
                             showlegend=show_legend,
                             hoverlabel=hoverlabel,
                             legendgroup=legend_group,
                             legendgrouptitle_text=legend_group_title,
                             **kwargs)



import sys

# import the actual implementation package
import uplt

# import everything is do nothing really
# only for static type checkers and analysis tools
from uplt import * # type: ignore # noqa: F403

# make 'uplot' a true alias for 'uplt' in the module cache
sys.modules['uplot'] = uplt



import numpy as np
from typing import Literal, cast

from uplt.interface import AxisScale


def get_scale(figure, axis: Literal['x', 'y', 'z']) -> AxisScale:
    axis_key = { 'x': 'xaxis', 'y': 'yaxis' }[axis]
    axis_type = getattr(getattr(figure.layout, axis_key), 'type', 'linear')
    axis_type = cast(AxisScale, axis_type)
    return axis_type


def set_scale(figure, axis: Literal['x', 'y'], scale: AxisScale, base: float):
    if axis == 'x':
        update_axes = figure.update_xaxes
    elif axis == 'y':
        update_axes = figure.update_yaxes
    else:
        raise ValueError(f'unsupported axis: {axis}')

    update_axes(type=scale)
    if scale == 'log':
        update_axes(dtick=np.log10(base))
    elif scale == 'linear':
        update_axes(dtick=None)
    else:
        raise ValueError(f'unsupported scale: {scale}')



import uplt
import imageio.v3 as iio # type: ignore

## Run Mode

save_to_file = False
engine = 'mpl'

## Data

x = [4, 5, 6, 8]

y1 = [1, 2, 3, 4]
y2 = [4, 3, 2, 1]

## Fig 1

f1 = uplt.figure(engine=engine)
f1.plot(x, y1)
f1.plot(x, y2, line_style=':', color='gray')
f1.scatter(3, 5, name='The lonely dot')
f1.title('Test')
f1.legend(True)
f1.axis_aspect('equal')
f1.xlabel('X value')
f1.ylabel('Y value')

if save_to_file:
    f1.save('fig1.save.png')
    iio.imwrite('fig1.as_image.png', f1.as_image())
else:
    f1.show(block=False)

## Fig 2

f2 = uplt.figure(engine=engine, aspect_ratio=0.4)
f2.plot(y1, x, name='#1', line_style='--')
f2.plot(y2, x)
f2.title('Test')
f2.legend(True)
f2.xlabel('X value')
f2.ylabel('Y value')
f2.xlim(-0.5)
f2.ylim(0)

if save_to_file:
    f2.save('fig2.save.png')
    iio.imwrite('fig2.as_image.png', f2.as_image())
else:
    f2.show()



import uplt.engine as uengine
from uplt.interface import IFigure, IPlotEngine
from uplt.default import DEFAULT

# cache variable to store last used engine
CURRENT_ENGINE: IPlotEngine | None = None


def figure(engine      : str | IPlotEngine | None = None,
           width       : int = DEFAULT.figure_width,
           aspect_ratio: float = DEFAULT.figure_aspect_ratio) -> IFigure:
    """
    Create a new figure using the specified plotting engine or the default engine.

    Parameters
    ----------
    engine : str, IPlotEngine, or None, optional
        The plotting engine object, name, or None.
        If None, the function uses the previous or default engine.

    width : int or None, optional
        The width of the figure in pixels.

    aspect_ratio : float, optional
        The aspect ratio for the new figure. The value in the range (0, 4].

    Returns
    -------
    IFigure
        The created figure object.

    Examples
    --------
    >>> fig = figure(engine='plotly', width=800, aspect_ratio=0.6)
    >>> fig.plot(x, y, name='Data')
    >>> fig.show()

    >>> fig = figure(width=1200)
    >>> fig.scatter(x, y)
    >>> fig.show()
    """
    global CURRENT_ENGINE

    if not 0 < aspect_ratio <= 4:
        raise ValueError('Aspect ratio must be in the range (0, 4].')

    if engine is None:
        if CURRENT_ENGINE is not None:
            # use the previously used engine
            engine = CURRENT_ENGINE
        else:
            # use the first available engine
            engine = uengine.get()
            if engine is None:
                raise RuntimeError('No plotting engines are registered.')

    elif isinstance(engine, str):
        # get the engine object by name
        engine = uengine.get(engine)
        if engine is None:
            raise RuntimeError(f'Plotting engine "{engine}" is not registered.')

    return engine.figure(width=width, aspect_ratio=aspect_ratio)



from uplt.interface import IPlotEngine
from typing import OrderedDict


DEFAULT_ENGINES: dict[str, IPlotEngine] = OrderedDict[str, IPlotEngine]()


def available() -> dict[str, list[str]]:
    """
    Retrieve a dictionary mapping available plot engines to their registered shortcut names.

    Returns
    -------
    dict[str, list[str]]
        A mapping between plot engines and their corresponding registered names.
    """
    engine_name_mapping = {}

    for name, engine in DEFAULT_ENGINES.items():
        if engine.name not in engine_name_mapping:
            # create a list of names for the engine
            engine_name_mapping[engine.name] = []
        # add the registered name for the engine
        engine_name_mapping[engine.name].append(name)

    return engine_name_mapping


def register(engine: IPlotEngine, name: str) -> bool:
    """
    Register a plot engine object with a specified shortcut name for use in the **figure()** function.

    Parameters
    ----------
    engine : IPlotEngine
        The plot engine object to register.

    name : str
        The shortcut name for the engine.

    Returns
    -------
    bool
        Returns True if the registration is successful.
        Returns False if the engine is unavailable or the name is already occupied.
    """
    name = name.lower()

    if name in DEFAULT_ENGINES:
        return False

    if not engine.is_available():
        return False

    DEFAULT_ENGINES[name] = engine
    return True


def get(name: str | None = None) -> IPlotEngine | None:
    """
    Retrieve a registered plot engine object by its specified name.

    Parameters
    ----------
    name : str or None, optional
        The name of the plot engine. If None, the first registered engine will be returned.

    Returns
    -------
    IPlotEngine or None
        The plot engine object corresponding to the specified name, or None if not found.
    """
    if len(DEFAULT_ENGINES) == 0:
        raise RuntimeError('No plotting engines are registered.')

    if name is None:
        # return the first available engine
        return list(DEFAULT_ENGINES.values())[0]

    engine = DEFAULT_ENGINES.get(name.lower())
    return engine



from __future__ import annotations

import numpy as np
from pathlib import Path
from numpy import ndarray
from typing import Any
from typing_extensions import override # Python 3.12 feature
from numpy.typing import ArrayLike
from collections.abc import Sequence

import uplt.color as ucolor
import uplt.utool as utool
import uplt.plugin as plugin
import uplt.detect as detect

from uplt.default import DEFAULT
from uplt.interface import IFigure
from uplt.utool import Interpolator
from uplt.engine.MatplotEngine import MatplotEngine
from uplt.interface import Colormap, ColormapMode
from uplt.interface import LineStyle, MarkerStyle, AspectMode, AxisScale


class MatplotFigure(IFigure):
    # defaults
    SHOWING_DPI = 100
    SAVING_DPI = SHOWING_DPI * 2
    LEGEND_MARKER_SIZE = 8

    @property
    @override
    def engine(self) -> MatplotEngine:
        return self._engine

    @property
    @override
    def internal(self) -> Any | None:
        return self._fig

    @property
    @override
    def is_3d(self) -> bool:
        assert self._is_3d is not None, 'axis must be initialized'
        return self._is_3d


    def __init__(self, engine: MatplotEngine, width: int, aspect_ratio: float):
        # temporary styling (no global effect):
        # https://matplotlib.org/stable/users/explain/customizing.html
        with engine.plt.style.context(DEFAULT.style):
            self._fig = engine.plt.figure(dpi=self.SHOWING_DPI, layout='constrained')
            # constrained layout automatically adjusts subplots so that decorations like tick labels,
            # legends, and colorbars do not overlap, while still preserving the logical layout requested by the user.
            # constrained layout is similar to Tight layout, but is substantially more flexible.
            # https://matplotlib.org/stable/users/explain/axes/constrainedlayout_guide.html
            self._fig.set_figwidth(width / self.SHOWING_DPI)
            self._fig.set_figheight(aspect_ratio*(width / self.SHOWING_DPI))

        self._engine = engine
        self._color_scroller = ucolor.ColorScroller()
        self._is_3d = None
        self._init_axis(is_3d=False)
        assert self._axis is not None
        self._bars = [ ]


    @override
    def plot(self, x           : ArrayLike,
                   y           : ArrayLike | None = None,
                   z           : ArrayLike | None = None,
                   name        : str | None = None,
                   color       : str | Sequence[str] | None = None,
                   line_style  : LineStyle | None = None,
                   marker_style: MarkerStyle | None = None,
                   marker_size : float | None = None,
                   opacity     : float = 1.0,
                   legend_group: str | None = None,
                   **kwargs) -> IFigure:
        from uplt.engine.matplot.plot import plot_line_marker

        # check if x is a custom object and a plugin is available
        if plugin.plot(plot_method=self.plot,
                       x=x, y=y, z=z,
                       name=name,
                       color=color,
                       line_style=line_style,
                       marker_style=marker_style,
                       marker_size=marker_size,
                       opacity=opacity,
                       legend_group=legend_group,
                       **kwargs):
            return self

        # get or init axis
        axis = self._init_axis(is_3d=z is not None)

        # init color
        assert isinstance(color, str | None), 'color must be a string or None for line plot'
        if color is None:
            color = self.scroll_color()

        plot_line_marker(axis=axis,
                         x=x, y=y, z=z,
                         name=name,
                         color=color,
                         line_style=line_style,
                         marker_style=marker_style,
                         marker_size=marker_size,
                         opacity=opacity,
                         **kwargs)
        return self


    @override
    def scatter(self, x           : ArrayLike,
                      y           : ArrayLike | None = None,
                      z           : ArrayLike | None = None,
                      name        : str | None = None,
                      color       : str | Sequence[str] | None = None,
                      marker_style: MarkerStyle | None = None,
                      marker_size : float | None = None,
                      opacity     : float = 1.0,
                      legend_group: str | None = None,
                      **kwargs) -> IFigure:
        from uplt.engine.matplot.plot import plot_line_marker

        # check if x is a custom object and a plugin is available
        if plugin.plot(plot_method=self.scatter,
                       x=x, y=y, z=z,
                       name=name,
                       color=color,
                       marker_style=marker_style,
                       marker_size=marker_size,
                       opacity=opacity,
                       legend_group=legend_group,
                       **kwargs):
            return self

        # get or init axis
        axis = self._init_axis(is_3d=z is not None)

        # init color
        if color is None:
            color = self.scroll_color()

        plot_line_marker(axis=axis,
                         x=x, y=y, z=z,
                         name=name,
                         color=color,
                         line_style=' ',  # no line (scatter mode)
                         marker_style=marker_style,
                         marker_size=marker_size,
                         opacity=opacity,
                         **kwargs)
        return self


    @override
    def hline(self, y           : float,
                    x_min       : float | None = None,
                    x_max       : float | None = None,
                    name        : str | None = None,
                    color       : str | None = None,
                    line_style  : LineStyle | None = None,
                    opacity     : float = 1.0,
                    legend_group: str | None = None,
                    **kwargs) -> IFigure:

        if self.is_3d:
            raise RuntimeError('3D figure is not supported')

        if x_min is None:
            x_min, _ = self._axis.get_xlim()

        if x_max is None:
            _, x_max = self._axis.get_xlim()

        return self.plot([x_min, x_max], [y, y],
                         color=color,
                         name=name,
                         line_style=line_style,
                         opacity=opacity,
                         legend_group=legend_group,
                         **kwargs)


    @override
    def vline(self, x           : float,
                    y_min       : float | None = None,
                    y_max       : float | None = None,
                    name        : str | None = None,
                    color       : str | None = None,
                    line_style  : LineStyle | None = None,
                    opacity     : float = 1.0,
                    legend_group: str | None = None,
                    **kwargs) -> IFigure:
        if self.is_3d:
            raise RuntimeError('3D figure is not supported')

        if y_min is None:
            y_min, _ = self._axis.get_ylim()

        if y_max is None:
            _, y_max = self._axis.get_ylim()

        return self.plot([x, x], [y_min, y_max],
                         color=color,
                         name=name,
                         line_style=line_style,
                         opacity=opacity,
                         legend_group=legend_group,
                         **kwargs)


    @override
    def surface3d(self, x            : ArrayLike | Any,
                        y            : ArrayLike | None = None,
                        z            : ArrayLike | None = None,
                        name         : str | None = None,
                        show_colormap: bool = False,
                        colormap     : Colormap = 'viridis',
                        opacity      : float = 1.0,
                        interpolation: Interpolator = 'cubic',
                        interpolation_range: int = 100,
                        legend_group : str | None = None,
                        **kwargs) -> IFigure:
        assert self._fig is not None, 'figure is closed'

        # check if x is a custom object and a plugin is available
        if plugin.plot(plot_method=self.surface3d,
                       x=x, y=y, z=z,
                       name=name,
                       show_colormap=show_colormap,
                       colormap=colormap,
                       opacity=opacity,
                       interpolation=interpolation,
                       interpolation_range=interpolation_range,
                       legend_group=legend_group,
                       **kwargs):
            return self

        x = np.asarray(x)
        y = np.asarray(y)
        z = np.asarray(z)
        assert x.ndim == y.ndim == 1, 'x, y must be 1D arrays'
        assert z.ndim == 1 or z.ndim == 2, 'z must be 1D or 2D array'

        if z.ndim == 2:
            # uniform grid
            assert (len(y), len(x)) == z.shape, 'uniform grid: x and y range must match z'
            x, y = np.meshgrid(x, y)
        else:
            # non-uniform grid - array of points (x, y, z)
            x, y, z = utool.array_to_grid(x, y, z,
                                          interpolation=interpolation,
                                          interpolation_range=interpolation_range)

        axis = self._init_axis(is_3d=True)

        from mpl_toolkits.mplot3d import Axes3D
        assert isinstance(axis, Axes3D)

        cmap = self.engine.mpl.colormaps[colormap.lower()]

        surf = axis.plot_surface(x, y, z,
                                 label=name,
                                 cmap=cmap,
                                 # antialiased=False: to fix some of z-order issues (not all of them)
                                 # https://stackoverflow.com/questions/39144482/matplotlib-plot-surface-transparency-artefact
                                 antialiased=False,
                                 alpha=opacity,
                                 **kwargs)
        if show_colormap:
            self._fig.colorbar(surf, shrink=0.5, aspect=10)

        return self


    @override
    def bar(self, x           : ArrayLike,
                  y           : ArrayLike | None = None,
                  name        : str | None = None,
                  color       : str | None = None,
                  opacity     : float = 1.0,
                  legend_group: str | None = None,
                  **kwargs) -> IFigure:
        # get or init axis
        axis = self._init_axis(is_3d=False)

        x = np.asarray(x)
        if y is None:
            # y is provided via x
            y = x
            x = np.arange(len(y))
        else:
            y = np.asarray(y)
            assert len(x) == len(y), 'the length of the input arrays must be the same'

        # init color
        if color is None:
            color = self.scroll_color()

        # update bar params
        total_width = kwargs.pop('width', 0.8)
        bar_ofs = (1 - total_width) / 2

        bar_idx = len(self._bars)
        bar_width = total_width / (bar_idx + 1)

        # adjust pos and widths of existing bars
        for idx, bar_i in enumerate(self._bars):
            x_idx = np.arange(len(bar_i.patches))
            bar_pos = x_idx + bar_ofs + idx*bar_width
            for j in range(len(bar_i.patches)):
                bar_i.patches[j].set_x(bar_pos[j] - bar_width/2)
                bar_i.patches[j].set_width(bar_width)

        # add new bar
        x_idx = np.arange(len(y))
        bar_pos = x_idx + bar_ofs + bar_idx*bar_width

        rects = axis.bar(bar_pos, y,
                         width=bar_width,
                         color=ucolor.name_to_hex(color),
                         alpha=opacity,
                         label=name,
                         **kwargs)

        # set ticks names: str or numbers
        axis.set_xticks(x_idx - bar_width/2 + 0.5, x)

        # save bar for future adjustments
        self._bars.append(rects)

        return self


    @override
    def imshow(self, image: ArrayLike, cmap: Colormap | None = None, **kwargs) -> IFigure:
        image = np.asarray(image)

        if 'vmin' in kwargs or 'vmax' in kwargs:
            # fallback to matplotlib behaviour if
            # the image range provided directly
            vmin = kwargs.pop('vmin', None)
            vmax = kwargs.pop('vmax', None)
        else:
            # test the image for type and convert to [0, 1] range
            image = image / utool.image_range(image)
            vmin = 0.0
            vmax = 1.0

        axis = self._init_axis(is_3d=False)
        cmap = cmap or 'gray'
        axis.imshow(image,
            cmap=cmap,
            vmin=vmin, vmax=vmax,
            interpolation=kwargs.pop('interpolation', 'none')
        )

        # hide grid, frame, ticks and labels
        axis.grid(visible=False)
        axis.get_xaxis().set_visible(False)
        axis.get_yaxis().set_visible(False)
        axis.set_frame_on(False)

        return self


    @override
    def heatmap(self, data    : ArrayLike,
                      cmap    : Colormap = 'jet',
                      colorbar: ColormapMode = 'vertical') -> IFigure:
        data = np.asarray(data)
        assert data.ndim == 2 or data.shape[2] == 1, \
               'heatmap data must be 2D array or 3D array with one channel'

        axis = self._init_axis(is_3d=False)
        fig = self._fig
        assert fig is not None

        img = axis.imshow(data, cmap=cmap, vmin=data.min(), vmax=data.max())

        axis.grid(visible=False) # hide grid
        axis.xaxis.tick_top()    # move x-axis to top

        if colorbar != 'off':
            fig.colorbar(img, orientation=colorbar, pad=0.01)

        return self


    @override
    def title(self, text: str) -> IFigure:
        self._axis.set_title(label=text)
        return self


    @override
    def legend(self, show: bool = True,
                     equal_marker_size: bool = True,
                     **kwargs) -> IFigure:
        assert self._fig is not None, 'figure is closed'

        if not self._axis:
            return self

        # check if there is anything to put to the legend
        handles, labels = self._axis.get_legend_handles_labels()
        if len(handles) == 0:
            return self

        if equal_marker_size:
            from matplotlib.legend_handler import HandlerPathCollection, HandlerLine2D
            from matplotlib.collections import PathCollection
            from matplotlib.lines import Line2D

            def updatescatter(handle, orig):
                handle.update_from(orig)
                handle.set_sizes([self.LEGEND_MARKER_SIZE ** 2])
            def updateline(handle, orig):
                handle.update_from(orig)
                handle.set_markersize(self.LEGEND_MARKER_SIZE)

            handler_map = { PathCollection: HandlerPathCollection(update_func=updatescatter),
                            Line2D: HandlerLine2D(update_func=updateline) }
        else:
            handler_map = None

        if not show:
            self._fig.legends.clear() # remove legend outside axis
            self._fig.gca().legend().remove() # remove legend inside
            return self

        # create legend
        loc = kwargs.pop('loc', 'outside right upper')
        if 'outside' in loc:
            # outside works only for the figure
            # "outside right upper" works correctly with "constrained" or "compressed" layout only
            self._fig.legend(handler_map=handler_map).set(loc=loc, **kwargs)
        else:
            # axes.legend() is better for an other options because legend will be inside the graph
            self._fig.gca().legend(handler_map=handler_map).set(loc=loc, **kwargs)

        return self


    @override
    def grid(self, show: bool = True) -> IFigure:
        self._axis.grid(visible=show, which='both')
        return self


    @override
    def xlabel(self, text: str) -> IFigure:
        self._axis.set_xlabel(xlabel=text)
        return self


    @override
    def ylabel(self, text: str) -> IFigure:
        self._axis.set_ylabel(ylabel=text)
        return self


    @override
    def zlabel(self, text: str) -> IFigure:
        if self.is_3d:
            from mpl_toolkits.mplot3d import Axes3D
            assert isinstance(self._axis, Axes3D)
            self._axis.set_zlabel(zlabel=text)
        return self


    @override
    def xlim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        self._axis.set_xlim(left=min_value, right=max_value)
        return self


    @override
    def ylim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        self._axis.set_ylim(bottom=min_value, top=max_value)
        return self


    @override
    def zlim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        if self.is_3d:
            from mpl_toolkits.mplot3d import Axes3D
            assert isinstance(self._axis, Axes3D)
            self._axis.set_zlim(bottom=min_value, top=max_value)
        return self


    @override
    def xscale(self, scale: AxisScale, base: float = 10) -> IFigure:
        if scale == 'linear':
            self._axis.set_xscale(scale)
        elif scale == 'log':
            self._axis.set_xscale(scale, base=base)
        return self


    @override
    def yscale(self, scale: AxisScale, base: float = 10) -> IFigure:
        if scale == 'linear':
            self._axis.set_yscale(scale)
        elif scale == 'log':
            self._axis.set_yscale(scale, base=base)
        return self


    @override
    def current_color(self) -> str:
        return self._color_scroller.current_color()


    @override
    def scroll_color(self, count: int=1) -> str:
        return self._color_scroller.scroll_color(count)


    @override
    def reset_color(self) -> IFigure:
        self._color_scroller.reset()
        return self


    @override
    def axis_aspect(self, mode: AspectMode) -> IFigure:
        # https://stackoverflow.com/questions/8130823/set-matplotlib-3d-plot-aspect-ratio
        self._axis.set_aspect(aspect=mode)
        return self


    @override
    def as_image(self) -> ndarray:
        if self._fig is None:
            raise RuntimeError('figure is closed')

        fig = self._fig

        fig.set_dpi(self.SAVING_DPI)
        fig.canvas.draw()

        from matplotlib.backends.backend_agg import FigureCanvasAgg
        assert isinstance(fig.canvas, FigureCanvasAgg)

        image = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)

        w, h = fig.canvas.get_width_height()
        return image.reshape([h, w, 4])


    @override
    def save(self, filename: str | Path) -> IFigure:
        assert self._fig is not None, 'figure is closed'
        self._fig.savefig(filename, dpi=self.SAVING_DPI)
        return self


    @override
    def close(self):
        self.engine.plt.close(self._fig)
        self._fig = None


    @override
    def show(self, block: bool=True):
        assert self._fig is not None, 'figure is closed'

        if detect.is_marimo():
            # marimo can visualize matplotlib figures directly
            return self.internal

        if self.engine.is_ipython_backend:
            # there are two ways for consistent figure visualization in jupyter
            #    1. call `%matplotlib ...` at the notebook start.
            #    2. always use `plt.show()`.
            # it's easy to forget about (1) so `plt.show()` is more reliable
            # but not the best, probably (because it will show all figures).
            self.engine.plt.show()
            return

        if not self.engine.is_gui_backend:
            return # no need to show, bypass

        self._fig.show() # show only this figure

        if block:
            while self.engine.plt.fignum_exists(self._fig.number):
                # allow multiple mouse clicks for 3d plot manipulation
                self._fig.waitforbuttonpress()


    ## Protected ##


    def _init_axis(self, is_3d: bool):
        assert self._fig is not None, 'figure is closed'

        if self._is_3d == is_3d:
            # axis already initialized
            return self._axis

        # remove current axis
        self._fig.clear()

        # temporary styling (no global effect):
        with self.engine.plt.style.context(DEFAULT.style):
            projection = '3d' if is_3d else None
            self._axis = self._fig.add_subplot(projection=projection)

        self._is_3d = is_3d
        self._axis.grid(visible=True, which='both') # show grid by default

        if is_3d:
            # sync axis and figure color
            self._axis.set_facecolor(self._fig.get_facecolor())

        return self._axis



import importlib.util
from typing_extensions import override # Python 3.12 feature
from uplt.interface import IPlotEngine, IFigure
from uplt.default import DEFAULT


class PlotlyEngine5(IPlotEngine):

    @property
    @override
    def name(self) -> str:
        return 'plotly'

    @classmethod
    @override
    def is_available(cls) -> bool:
        return importlib.util.find_spec('plotly') is not None

    @property
    def go(self):
        return self._go

    @property
    def pio(self):
        return self._pio


    def __init__(self):
        import plotly.graph_objs as go
        import plotly.io as pio

        self._pio = pio
        self._go = go

        # load style
        if DEFAULT.style.lower() == 'bmh':
            from uplt.engine.style.plotly import bmh
            self._layout_style = bmh
        else:
            raise NotImplementedError(f'style not supported for plotly: {DEFAULT.style}')


    @override
    def figure(self, width: int, aspect_ratio: float) -> IFigure:
        from uplt.engine.PlotlyFigure5 import PlotlyFigure5
        fig = PlotlyFigure5(engine=self) # type: ignore

        # adjust style layout
        assert fig.internal is not None
        fig.internal.update_layout(template=self._layout_style,
                                   width=width,
                                   height=aspect_ratio*width)

        return fig



import marimo

__generated_with = "0.16.5"
app = marimo.App(width="medium")


@app.cell
def _():
    import marimo as mo
    return (mo,)


@app.cell
def _():
    import numpy as np
    return (np,)


@app.cell
def _():
    import uplt
    return (uplt,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# Test Notebook Detection""")
    return


@app.cell
def _():
    from uplt import detect
    return (detect,)


@app.cell
def _(detect):
    detect.is_marimo()
    return


@app.cell
def _(detect):
    detect.is_jupyter()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# Data""")
    return


@app.cell
def _(np):
    x = np.linspace(0, np.pi*4, num=100)
    y1 = np.sin(x)
    y2 = np.sin(x - np.pi/4)
    y3 = np.sin(x - 2*np.pi/4)
    y4 = np.sin(x - 3*np.pi/4)
    return x, y1, y2, y3, y4


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# Matplotlib""")
    return


@app.cell
def _(mo, uplt, x, y1, y2, y3, y4):
    mpl_fig = (
        uplt.figure('mpl')
        .plot(x, y1, name='data #1')
        .plot(x, y2, name='data #2')
        .plot(x, y3, name='data #3')
        .plot(x, y4, name='data #4')
        .xlabel('X')
        .ylabel('Y')
        .legend()
    )

    mo.md(f'**Matplotlib Image**: {mo.as_html(mpl_fig.show())}').center()
    return


@app.cell
def _(uplt, x, y1, y2, y3, y4):
    (
        uplt.figure('mpl')
        .plot(x, y1, name='data #1')
        .plot(x, y2, name='data #2')
        .plot(x, y3, name='data #3')
        .plot(x, y4, name='data #4')
        .xlabel('X')
        .ylabel('Y')
        .legend()
    ).show()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# Plotly""")
    return


@app.cell
def _(mo, uplt, x, y1, y2, y3, y4):
    plt_fig = (
        uplt.figure('plotly')
        .plot(x, y1, name='data #1')
        .plot(x, y2, name='data #2')
        .plot(x, y3, name='data #3')
        .plot(x, y4, name='data #4')
        .xlabel('X')
        .ylabel('Y')
        .legend()
    )

    mo.md(f'**Plotly Image**: {mo.as_html(plt_fig.show())}').center()
    return


@app.cell
def _(uplt, x, y1, y2, y3, y4):
    (
        uplt.figure('plotly')
        .plot(x, y1, name='data #1')
        .plot(x, y2, name='data #2')
        .plot(x, y3, name='data #3')
        .plot(x, y4, name='data #4')
        .xlabel('X')
        .ylabel('Y')
        .legend()
    ).show()
    return


if __name__ == "__main__":
    app.run()



from dataclasses import dataclass


@dataclass
class Default:
    figure_width       : int = 800
    figure_aspect_ratio: float = 0.6
    style              : str = 'bmh'
    marker_size        : int = 6


DEFAULT = Default()


import numpy as np
from numpy import ndarray


def image_range(image: ndarray) -> int | float:
    """
    Guess the range by an image type and values
    """
    # int image
    if image.dtype.type == np.uint8:
        return 255

    if image.dtype.type == np.uint16:
        return 65535

    # float image
    max_value = np.max(image)

    if max_value < 1.01:
        return 1.0

    if max_value < 255:
        return 255

    if max_value < 65535:
        return 65535

    raise RuntimeError('image range detection failure')


def image_encode_base64(image: ndarray, value_range: float) -> str:
    import base64
    from io import BytesIO
    from PIL import Image

    image_u8 = (255 * (image / value_range)).astype(np.uint8)

    pil_image = Image.fromarray(image_u8)
    prefix = 'data:image/png;base64,'

    with BytesIO() as stream:
        pil_image.save(stream, format='png')
        image_base64 = prefix + base64.b64encode(stream.getvalue()).decode('utf-8')

    return image_base64


import uplt

import numpy as np
import pandas as pd # type: ignore

import uplt.plugin as plugin


class DataFramePlugin(plugin.IPlotPlugin):
    """
    DataFrame minimalistic plugin (only data extraction)
    """

    def extract_data(self, obj: pd.DataFrame) -> list[plugin.PlotData]:
        data = []
        for name in obj.columns:
            if not np.issubdtype(obj.dtypes[name], np.number):
                continue
            y = np.asarray(obj[name].values)
            x = np.arange(len(y))
            name = name.replace('_', ' ').title()
            data.append(plugin.PlotData(x=x, y=y, name=name))
        return data


class DataFramePluginAdvanced(plugin.IPlotPlugin):
    """
    DataFrame plugin with advanced style processing
    """

    def __init__(self):
        # joint name for all data in DataFrame
        self._joined_name = None


    def extract_data(self, obj: pd.DataFrame) -> list[plugin.PlotData]:
        data = []
        joined_name = []
        for name in obj.columns:
            if not np.issubdtype(obj.dtypes[name], np.number):
                continue
            y = np.asarray(obj[name].values)
            x = np.arange(len(y))
            name = name.replace('_', ' ').title()
            joined_name.append(name)
            data.append(plugin.PlotData(x=x, y=y, name=name))
        self._joined_name = ' | '.join(joined_name)
        return data


    def update_style(self, plot_type : plugin.PlotType,
                           data_index: int,
                           data_count: int,
                           data_name : str | None,
                           group_name: str | None,
                           **kwargs) -> dict:
        name = kwargs.get('name', None)
        if name is not None:
            return kwargs

        kwargs['name'] = data_name

        if kwargs.get('color', None) is not None:
            # single color mode:
            #   - the one legend entry will be shown for all datasets.
            #    (no reason to have multiple entries with a single color)
            #   - all names for the data will be combined into one.
            if data_index == data_count - 1:
                # set the combined name for the last dataset only
                kwargs['name'] = self._joined_name
            else:
                kwargs['name'] = None
            # combine everything into the same legend group (plotly)
            kwargs['legend_group'] = self._joined_name

        return kwargs


# plugin.register(pd.DataFrame, handler=DataFramePlugin())
plugin.register(pd.DataFrame, handler=DataFramePluginAdvanced())


car_crashes = pd.read_csv('https://raw.githubusercontent.com/mwaskom/seaborn-data/master/car_crashes.csv')


fig = uplt.figure('mpl')
# fig = uplot.figure('plotly')

# regular plot
fig.plot(car_crashes[['total', 'speeding', 'alcohol', 'no_previous']])

# single color mode
# fig.plot(car_crashes[['total', 'speeding', 'alcohol', 'no_previous']], color='green')

fig.legend()
fig.show()



import numpy as np
from typing import Literal


Interpolator = Literal[
    'nearest',
    'linear',
    'cubic',
]


def array_to_grid(x: np.ndarray,
                  y: np.ndarray,
                  z: np.ndarray,
                  interpolation: Interpolator,
                  interpolation_range: int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Convert an array of non-uniformly distributed points to a uniformly interpolated grid.

    parameters
    ----------
    x, y : np.ndarray
        1D arrays of coordinates.
    z : np.ndarray
        1D array of corresponding values for (x, y).
    interpolation : Interpolator
        Interpolation method to be used ('nearest', 'linear', or 'cubic').
    interpolation_range : int
        Number of points for creating the uniform grid.

    returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        x, y, z - 2D arrays representing uniform grid coordinates and a 2D array of corresponding values.
    """
    assert len(x) == len(y) == len(z), 'sizes of x, y, z must match'
    from scipy.interpolate import griddata

    # create a uniform grid based on the specified interpolation range
    x_range = np.linspace(x.min(), x.max(), num=interpolation_range)
    y_range = np.linspace(y.min(), y.max(), num=interpolation_range)

    # perform interpolation to obtain the uniformly distributed grid
    z = griddata((x, y), z, (x_range[None, :], y_range[:, None]), method=interpolation)
    x, y = np.meshgrid(x_range, y_range)

    return x, y, z


from types import GenericAlias
from uplt.plugin.IPlotPlugin import IPlotPlugin


REGISTERED_TYPES: dict[type | GenericAlias, IPlotPlugin] = {}


def is_registered(t: type | GenericAlias) -> bool:
    """
    Checks whether type **t** is registered and has a handler.
    """
    return t in REGISTERED_TYPES


def register(t: type | GenericAlias, handler: IPlotPlugin, force: bool = False) -> bool:
    """
    Register a handler for the specified type **t**.

    Parameters
    ----------
    t : type
        The type to register.

    handler : IPlotPlugin
        The plugin for data extraction from the specified type **t**.

    force : bool, optional
        If True, set the handler even if a handler is already registered for the type.

    Returns
    -------
    bool
        True if registration is successful, False otherwise.
    """
    if isinstance(handler, type):
        raise RuntimeError('handler must be instance not a type')

    if is_registered(t) and not force:
        # already registered
        return False

    REGISTERED_TYPES[t] = handler
    return True


def get_handler(t: type | GenericAlias) -> IPlotPlugin | None:
    """
    Get plotting plugin(handler) for the registered type.
    """
    if is_registered(t):
        return REGISTERED_TYPES[t]

    return None



"""
Utility functions to detect the current environment: jupyter, marimo, etc.
"""


def is_jupyter() -> bool:
    """
    Check if the current environment is a Jupyter notebook.
    """
    try:
        from IPython.core.getipython import get_ipython # type: ignore

        shell = get_ipython().__class__.__name__

        if shell == 'ZMQInteractiveShell':
            # Jupyter notebook or qtconsole
            return True

    except Exception:
        pass

    return False


def is_marimo() -> bool:
    """
    Check if the current environment is a Marimo notebook.
    """
    try:
        import marimo # type: ignore
        return marimo.running_in_notebook()
    except Exception:
        pass
    return False



from __future__ import annotations

import numpy as np
from pathlib import Path
from numpy import ndarray
from typing import Any
from typing_extensions import override # Python 3.12 feature
from numpy.typing import ArrayLike
from collections.abc import Sequence

import uplt.color as ucolor
import uplt.utool as utool
import uplt.plugin as plugin
import uplt.detect as detect

from uplt.interface import IFigure
from uplt.utool import Interpolator
from uplt.engine.PlotlyEngine5 import PlotlyEngine5
from uplt.interface import Colormap, ColormapMode
from uplt.interface import LineStyle, MarkerStyle, AspectMode, AxisScale


class PlotlyFigure5(IFigure):
    # defaults
    FILE_RESOLUTION_SCALE = 2
    LINE_WIDTH = 2.5

    @property
    @override
    def engine(self) -> PlotlyEngine5:
        return self._engine

    @property
    @override
    def internal(self) -> Any | None:
        return self._fig

    @property
    @override
    def is_3d(self) -> bool:
        return self._is_3d


    def __init__(self, engine: PlotlyEngine5):
        self._engine = engine
        self._color_scroller = ucolor.ColorScroller()

        self._fig = engine.go.Figure()
        self._is_3d = False
        self._colorbar_x_pos = 1.0
        self._show_grid = True

        self._group_counter: dict[str | None, int] = { None: 0 }


    @override
    def plot(self, x           : ArrayLike,
                   y           : ArrayLike | None = None,
                   z           : ArrayLike | None = None,
                   name        : str | None = None,
                   color       : str | Sequence[str] | None = None,
                   line_style  : LineStyle | None = None,
                   marker_style: MarkerStyle | None = None,
                   marker_size : float | None = None,
                   opacity     : float = 1.0,
                   legend_group: str | None = None,
                   **kwargs) -> IFigure:
        from uplt.engine.plotly.plot import plot_line_marker

        # check if x is a custom object and a plugin is available
        if plugin.plot(plot_method=self.plot,
                       x=x, y=y, z=z,
                       name=name,
                       color=color,
                       line_style=line_style,
                       marker_style=marker_style,
                       marker_size=marker_size,
                       opacity=opacity,
                       legend_group=legend_group,
                       **kwargs):
            return self

        self._is_3d = z is not None

        assert isinstance(color, str | None), 'color must be a string or None for line plot'
        if color is None:
            color = self.scroll_color()

        self._update_group_counter(plot_name=name, legend_group=legend_group)

        plot_line_marker(figure=self._fig,
                         x=x, y=y, z=z,
                         color=color,
                         name=name,
                         line_style=line_style,
                         line_width=self.LINE_WIDTH,
                         marker_style=marker_style,
                         marker_size=marker_size,
                         opacity=opacity,
                         legend_group=legend_group,
                         legend_group_title=legend_group if self._group_counter[legend_group] > 0 else None,
                         **kwargs)
        return self


    @override
    def scatter(self, x           : ArrayLike,
                      y           : ArrayLike | None = None,
                      z           : ArrayLike | None = None,
                      name        : str | None = None,
                      color       : str | Sequence[str] | None = None,
                      marker_style: MarkerStyle | None = None,
                      marker_size : float | None = None,
                      opacity     : float = 1.0,
                      legend_group: str | None = None,
                      **kwargs) -> IFigure:
        from uplt.engine.plotly.plot import plot_line_marker

        # check if x is a custom object and a plugin is available
        if plugin.plot(plot_method=self.scatter,
                       x=x, y=y, z=z,
                       name=name,
                       color=color,
                       marker_style=marker_style,
                       marker_size=marker_size,
                       opacity=opacity,
                       legend_group=legend_group,
                       **kwargs):
            return self

        self._is_3d = z is not None

        if color is None:
            color = self.scroll_color()

        self._update_group_counter(plot_name=name, legend_group=legend_group)

        plot_line_marker(figure=self._fig,
                         x=x, y=y, z=z,
                         color=color,
                         name=name,
                         line_style=' ', # no line (scatter mode)
                         line_width=self.LINE_WIDTH,
                         marker_style=marker_style,
                         marker_size=marker_size,
                         opacity=opacity,
                         legend_group=legend_group,
                         legend_group_title=legend_group if self._group_counter[legend_group] > 0 else None,
                         **kwargs)
        return self


    @override
    def hline(self, y           : float,
                    x_min       : float | None = None,
                    x_max       : float | None = None,
                    name        : str | None = None,
                    color       : str | None = None,
                    line_style  : LineStyle | None = None,
                    opacity     : float = 1.0,
                    legend_group: str | None = None,
                    **kwargs) -> IFigure:

        if self.is_3d:
            raise RuntimeError('3D figure is not supported')

        if x_min is None:
            from uplt.engine.plotly.axis_range import estimate_axis_range
            x_min = estimate_axis_range(self._fig, axis='x', mode='min')

        if x_max is None:
            from uplt.engine.plotly.axis_range import estimate_axis_range
            x_max = estimate_axis_range(self._fig, axis='x', mode='max')

        return self.plot([x_min, x_max], [y, y],
                         color=color,
                         name=name,
                         line_style=line_style,
                         opacity=opacity,
                         legend_group=legend_group,
                         **kwargs)


    @override
    def vline(self, x           : float,
                    y_min       : float | None = None,
                    y_max       : float | None = None,
                    name        : str | None = None,
                    color       : str | None = None,
                    line_style  : LineStyle | None = None,
                    opacity     : float = 1.0,
                    legend_group: str | None = None,
                    **kwargs) -> IFigure:
        if self.is_3d:
            raise RuntimeError('3D figure is not supported')

        if y_min is None:
            from uplt.engine.plotly.axis_range import estimate_axis_range
            y_min = estimate_axis_range(self._fig, axis='y', mode='min')

        if y_max is None:
            from uplt.engine.plotly.axis_range import estimate_axis_range
            y_max = estimate_axis_range(self._fig, axis='y', mode='max')

        return self.plot([x, x], [y_min, y_max],
                         color=color,
                         name=name,
                         line_style=line_style,
                         opacity=opacity,
                         legend_group=legend_group,
                         **kwargs)


    @override
    def surface3d(self, x            : ArrayLike,
                        y            : ArrayLike | None = None,
                        z            : ArrayLike | None = None,
                        name         : str | None = None,
                        show_colormap: bool = False,
                        colormap     : Colormap = 'viridis',
                        opacity      : float = 1.0,
                        interpolation: Interpolator = 'cubic',
                        interpolation_range: int = 100,
                        legend_group : str | None = None,
                        **kwargs) -> IFigure:
        # check if x is a custom object and a plugin is available
        if plugin.plot(plot_method=self.surface3d,
                       x=x, y=y, z=z,
                       name=name,
                       show_colormap=show_colormap,
                       colormap=colormap,
                       opacity=opacity,
                       interpolation=interpolation,
                       interpolation_range=interpolation_range,
                       legend_group=legend_group,
                       **kwargs):
            return self

        x = np.asarray(x)
        y = np.asarray(y)
        z = np.asarray(z)
        assert x.ndim == y.ndim == 1, 'x, y must be 1D arrays'
        assert z.ndim == 1 or z.ndim == 2, 'z must be 1D or 2D array'

        if z.ndim == 2:
            # uniform grid
            assert (len(y), len(x)) == z.shape, 'uniform grid: x and y range must match z'
            x, y = np.meshgrid(x, y)
        else:
            # non-uniform grid - array of points (x, y, z)
            x, y, z = utool.array_to_grid(x, y, z,
                                          interpolation=interpolation,
                                          interpolation_range=interpolation_range)

        self._is_3d = True

        if show_colormap:
            colorbar = dict(len=0.5, x=self._colorbar_x_pos)
            self._colorbar_x_pos += 0.12
        else:
            colorbar = None

        self._update_group_counter(plot_name=name, legend_group=legend_group)

        self._fig.add_surface(x=x, y=y, z=z,
                              name=name,
                              showlegend=(name != '') and (name is not None),
                              showscale=show_colormap,
                              colorscale=colormap,
                              colorbar=colorbar,
                              opacity=opacity,
                              legendgroup=legend_group,
                              legendgrouptitle_text=legend_group if self._group_counter[legend_group] > 0 else None,
                              **kwargs)
        return self


    @override
    def bar(self, x           : ArrayLike,
                  y           : ArrayLike | None = None,
                  name        : str | None = None,
                  color       : str | None = None,
                  opacity     : float = 1.0,
                  legend_group: str | None = None,
                  **kwargs) -> IFigure:

        self._is_3d = False

        x = np.asarray(x)
        if y is None:
            # y is provided via x
            y = x
            x = np.arange(len(y))
        else:
            y = np.asarray(y)
            assert len(x) == len(y), 'the length of the input arrays must be the same'

        if color is None:
            color = self.scroll_color()

        self._update_group_counter(plot_name=name, legend_group=legend_group)

        if name is None:
            name = ''
            show_legend = False
        else:
            show_legend = kwargs.pop('showlegend', True)

        self._fig.add_bar(x=x, y=y,
                          marker_color=ucolor.name_to_hex(color),
                          name=name,
                          showlegend=show_legend,
                          legendgroup=legend_group,
                          opacity=opacity,
                          legendgrouptitle_text=legend_group if self._group_counter[legend_group] > 0 else None,
                          **kwargs)
        return self


    @override
    def imshow(self, image: ArrayLike, cmap: Colormap | None = None, **kwargs) -> IFigure:
        image = np.asarray(image)
        value_range = utool.image_range(image)

        fig = self._fig

        if image.ndim == 2 or image.shape[2] == 1:
            # Grayscale image workaround from plotly devs
            # https://github.com/plotly/plotly.py/issues/2885#issuecomment-724679904
            cmap = cmap or 'gray'
            fig.add_trace(self.engine.go.Heatmap(z=image, colorscale=cmap))
            fig.update_yaxes(autorange='reversed')   # origin at top-left corner
            fig.update_traces(dict(showscale=False)) # hide colorbar
            self.axis_aspect('equal')
        else:
            # Color image
            fig.add_trace(self.engine.go.Image(
                z=image,
                zmax=kwargs.pop('zmax', [value_range]*4),
                zmin=kwargs.pop('zmin', [0]*4),
                **kwargs,
            ))

        # Configure layout
        fig.update_layout(margin=self.engine.go.layout.Margin(b=30, t=30, pad=0))
        fig.update_layout(hovermode='closest')
        fig.update_xaxes(visible=False)
        fig.update_yaxes(visible=False)

        self._is_3d = False

        return self


    @override
    def heatmap(self, data    : ArrayLike,
                      cmap    : Colormap = 'jet',
                      colorbar: ColormapMode = 'vertical') -> IFigure:
        data = np.asarray(data)
        assert data.ndim == 2 or data.shape[2] == 1, \
               'heatmap data must be 2D array or 3D array with one channel'

        fig = self._fig

        if colorbar == 'vertical':
            cbar = dict(orientation='v')
        elif colorbar == 'horizontal':
            cbar = dict(orientation='h',
                        y=-0.15,          # Position below the plot (negative values)
                        xanchor='center', # Anchor point for x position
                        yanchor='top',    # Anchor point for y position
                        len=0.5)          # Length as fraction of plot width
        else:
            cbar = None

        fig.add_trace(
            self.engine.go.Heatmap(z=data, colorscale=cmap, colorbar=cbar)
        )

        self.axis_aspect('equal') # set equal aspect ratio for both axis
        fig.update_yaxes(autorange='reversed') # origin at top-left corner
        fig.update_layout(margin=self.engine.go.layout.Margin(b=30, t=50, pad=0))
        fig.update_layout(xaxis=dict(side='top')) # move x-axis to top
        fig.update_traces(dict(showscale = colorbar != 'off')) # show/hide colorbar

        return self


    @override
    def title(self, text: str) -> IFigure:
        self._fig.update_layout(title=text)
        return self


    @override
    def legend(self, show: bool = True,
                     equal_marker_size: bool = True,
                     **kwargs) -> IFigure:
        itemsizing = 'constant' if equal_marker_size else None

        self._fig.update_layout(legend=self.engine.go.layout.Legend(
            visible=show,
            bgcolor=kwargs.pop('bgcolor', 'rgba(255,255,255,0.8)'),
            itemsizing=kwargs.pop('itemsizing', itemsizing),
            itemwidth=kwargs.pop('itemwidth', 50),
            **kwargs,
        ))
        return self


    @override
    def grid(self, show: bool = True) -> IFigure:
        from uplt.engine.plotly.scale import get_scale

        show_minor_x = show and get_scale(self._fig, 'x') == 'log'
        show_minor_y = show and get_scale(self._fig, 'y') == 'log'
        if self.is_3d:
            Scene = self.engine.go.layout.Scene
            XAxis = self.engine.go.layout.scene.XAxis
            YAxis = self.engine.go.layout.scene.YAxis
            ZAxis = self.engine.go.layout.scene.ZAxis
            self._fig.update_layout(scene=Scene(xaxis=XAxis(showgrid=show),
                                                yaxis=YAxis(showgrid=show),
                                                zaxis=ZAxis(showgrid=show)))
        else:
            self._fig.update_xaxes(showgrid=show)
            self._fig.update_xaxes(minor=dict(ticks='inside' if show_minor_x else '',
                                              showgrid=show_minor_x))
            self._fig.update_yaxes(showgrid=show)
            self._fig.update_yaxes(minor=dict(ticks='inside' if show_minor_y else '',
                                              showgrid=show_minor_y))
        self._show_grid = show
        return self


    @override
    def xlabel(self, text: str) -> IFigure:
        if self.is_3d:
            self._fig.update_layout(scene=dict(xaxis_title=text))
        else:
            self._fig.update_xaxes(title=text)
        return self


    @override
    def ylabel(self, text: str) -> IFigure:
        if self.is_3d:
            self._fig.update_layout(scene=dict(yaxis_title=text))
        else:
            self._fig.update_yaxes(title=text)
        return self


    @override
    def zlabel(self, text: str) -> IFigure:
        if self.is_3d:
            self._fig.update_layout(scene=dict(zaxis_title=text))
        return self


    @override
    def xlim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        from uplt.engine.plotly.axis_range import estimate_axis_range
        from uplt.engine.plotly.scale import get_scale

        if min_value is None:
            min_value = estimate_axis_range(self._fig, axis='x', mode='min')

        if get_scale(self._fig, 'x') == 'log':
            min_value = np.log10(min_value)

        if max_value is None:
            max_value = estimate_axis_range(self._fig, axis='x', mode='max')

        if get_scale(self._fig, 'x') == 'log':
            max_value = np.log10(max_value)

        if self.is_3d:
            self._fig.update_layout(scene=dict(xaxis=dict(range=[min_value, max_value])))
        else:
            self._fig.update_xaxes(range=[min_value, max_value])
        return self


    @override
    def ylim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        from uplt.engine.plotly.axis_range import estimate_axis_range
        from uplt.engine.plotly.scale import get_scale

        if min_value is None:
            min_value = estimate_axis_range(self._fig, axis='y', mode='min')

        if get_scale(self._fig, 'y') == 'log':
            min_value = np.log10(min_value)

        if max_value is None:
            max_value = estimate_axis_range(self._fig, axis='y', mode='max')

        if get_scale(self._fig, 'y') == 'log':
            max_value = np.log10(max_value)

        if self.is_3d:
            self._fig.update_layout(scene=dict(yaxis=dict(range=[min_value, max_value])))
        else:
            self._fig.update_yaxes(range=[min_value, max_value])
        return self


    @override
    def zlim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        if not self.is_3d:
            return self

        if min_value is None:
            from uplt.engine.plotly.axis_range import estimate_axis_range
            min_value = estimate_axis_range(self._fig, axis='z', mode='min')

        if max_value is None:
            from uplt.engine.plotly.axis_range import estimate_axis_range
            max_value = estimate_axis_range(self._fig, axis='z', mode='max')

        self._fig.update_layout(scene=dict(zaxis=dict(range=[min_value, max_value])))
        return self


    @override
    def xscale(self, scale: AxisScale, base: float = 10) -> IFigure:
        from uplt.engine.plotly.scale import set_scale

        set_scale(self._fig, 'x', scale=scale, base=base)
        self.grid(self._show_grid) # update grid if visible
        return self


    @override
    def yscale(self, scale: AxisScale, base: float = 10) -> IFigure:
        from uplt.engine.plotly.scale import set_scale

        set_scale(self._fig, 'y', scale=scale, base=base)
        self.grid(self._show_grid) # update grid if visible
        return self


    @override
    def current_color(self) -> str:
        return self._color_scroller.current_color()


    @override
    def scroll_color(self, count: int=1) -> str:
        return self._color_scroller.scroll_color(count)


    @override
    def reset_color(self) -> IFigure:
        self._color_scroller.reset()
        return self


    @override
    def axis_aspect(self, mode: AspectMode) -> IFigure:
        if self.is_3d:
            aspectmode = 'cube' if mode == 'equal' else 'auto'
            self._fig.update_scenes(aspectmode=aspectmode)
        else:
            scaleanchor = 'x' if mode == 'equal' else None
            self._fig.update_yaxes(scaleanchor=scaleanchor, scaleratio=1)
            # Remove empty space around the plot when aspect ratio is equal
            self._fig.update_xaxes(constrain='domain')
            self._fig.update_yaxes(constrain='domain')
        return self


    @override
    def as_image(self) -> ndarray:
        import io
        from PIL import Image

        fig_bytes = io.BytesIO(
            self._fig.to_image(format='png', scale=self.FILE_RESOLUTION_SCALE)
        )

        image = Image.open(fig_bytes)
        image = np.asarray(image)
        image = image[..., :3] # RGBA -> RGB
        return image


    @override
    def save(self, filename: str | Path) -> IFigure:
        filename = Path(filename)
        if filename.suffix.lower() == '.html':
            self._fig.write_html(filename)
        else:
            self._fig.write_image(filename, scale=self.FILE_RESOLUTION_SCALE)
        return self


    @override
    def close(self):
        self._fig.data = []
        self._fig.layout = {}


    @override
    def show(self, block: bool=True):
        if detect.is_marimo():
            # marimo can visualize plotly figure directly
            return self.internal
        self.engine.pio.show(self._fig)


    ## Protected ##


    def _update_group_counter(self, plot_name: str | None, legend_group: str | None):
        """
        Count visible legend's items for the same group
        """
        if legend_group is None or len(legend_group) == 0:
            return

        group_size = self._group_counter.get(legend_group, 0)

        if plot_name is not None and len(plot_name) > 0:
            group_size += 1

        self._group_counter[legend_group] = group_size



from importlib.metadata import metadata

# engine managing
import uplt.engine as engine

# interface
from uplt.interface import IFigure, IPlotEngine

# main API function
from uplt.plot import figure

# common routines
import uplt.color as color

# common types
from uplt.utype import LineStyle, MarkerStyle, AspectMode, AxisScale, Colormap

# settings
from uplt.default import DEFAULT

# query the package metadata
meta = metadata('uplt-py')

__author__ = meta['Author-email']
__email__ = meta['Author-email']
__version__ = meta['Version']
__license__ = meta['License-Expression']


__all__ = [

    # package metadata

    "__version__",
    "__author__",
    "__email__",
    "__license__",

    # modules

    'engine',
    'color',

    # interface

    'IFigure',
    'IPlotEngine',

    # functions

    'figure',

    # types

    'LineStyle',
    'MarkerStyle',
    'AspectMode',
    'AxisScale',
    'Colormap',

    # variables / constants

    'DEFAULT'
]



import marimo

__generated_with = "0.18.3"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # Gallery
    """)
    return


@app.cell
def _():
    import uplt
    return (uplt,)


@app.cell(hide_code=True)
def _():
    import numpy as np
    return (np,)


@app.cell(hide_code=True)
def _():
    import marimo as mo
    return (mo,)


@app.cell(hide_code=True)
def _():
    from pathlib import Path
    return (Path,)


@app.cell(hide_code=True)
def _(mo):
    engine = mo.ui.dropdown(options=['matplot', 'plotly'],
                            value='matplot',
                            allow_select_none=False,
                            label='Engine:')
    engine
    return (engine,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # plot
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## 2d
    """)
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.linspace(0, np.pi*4, num=100)
    _y1_ = np.sin(_x_)
    _y2_ = np.sin(_x_ - np.pi/4)
    _y3_ = np.sin(_x_ - 2*np.pi/4)
    _y4_ = np.sin(_x_ - 3*np.pi/4)

    (
    uplt.figure(engine.value)
        .plot(_x_, _y1_, name='data #1')
        .plot(_x_, _y2_, name='data #2')
        .plot(_x_, _y3_, name='data #3')
        .plot(_x_, _y4_, name='data #4')
        .xlabel('X')
        .ylabel('Y')
        .legend()
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## 3d
    """)
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.linspace(0, np.pi*4, num=100)
    _y1_ = np.sin(_x_)
    _y2_ = np.sin(_x_ - np.pi/4)

    (
    uplt.figure(engine.value)
        .plot(x=_y1_, y=_y2_, z=_x_, name='data #1')
        .plot(x=_y1_, y=_y2_, z=_x_+2, name='data #2')
        .plot(x=_y1_, y=_y2_, z=_x_+4, name='data #1')
        .legend()
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # scatter
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## 2d
    """)
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.random.uniform(size=100)
    _y_ = np.random.uniform(size=100)


    uplt.figure(engine.value).scatter(_x_, _y_).show()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## 2d - Color per Point
    """)
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.random.uniform(size=100)
    _y_ = np.random.uniform(size=100)
    # random color per point
    _rgb_colors_ = np.random.uniform(low=0.1, high=0.8, size=[100, 3])


    uplt.figure(engine.value).scatter(_x_, _y_, color=uplt.color.rgb_to_str(_rgb_colors_)).show()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## 3d
    """)
    return


@app.cell
def _(engine, np, uplt):
    _z_ = np.linspace(0, np.pi*4, num=100)
    _x_ = np.sin(_z_)
    _y_ = np.sin(_z_ - np.pi/4)

    (
    uplt.figure(engine.value)
        .scatter(_x_, _y_, _z_, name='data #1')
        .scatter(_x_, _y_, _x_-5, name='data #2')
        .legend()
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## 3d - Color per Point
    """)
    return


@app.cell
def _(engine, np, uplt):
    _z_ = np.linspace(0, np.pi*4, num=100)
    _x_ = np.sin(_z_)
    _y_ = np.sin(_z_ - np.pi/4)

    _rgb_colors_ = np.random.uniform(low=0.1, high=0.9, size=[100, 3])


    uplt.figure(engine.value).scatter(_x_, _y_, _z_, color=uplt.color.rgb_to_str(_rgb_colors_)).show()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # surface3d
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Uniform Grid
    """)
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.arange(0, 5, 0.25)
    _y_ = np.arange(-5, 5, 0.25)
    _X_, _Y_ = np.meshgrid(_x_, _y_)
    _Z_ = np.sin(np.sqrt(_X_**2 + _Y_**2))

    (
    uplt.figure(engine.value, aspect_ratio=0.7)
        .surface3d(_x_, _y_, _Z_,   name='data #1', colormap='twilight')
        .surface3d(_x_, _y_, _Z_*6, name='data #2', show_colormap=True)
        .xlabel('X Axis')
        .ylabel('Y Axis')
        .zlabel('Z Axis')
        .legend()
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Non-uniform Grid
    """)
    return


@app.cell
def _(engine, uplt):
    # non-uniform grid (array of points)
    _x_ = [ 0, 1, 2, 0, 2, 1, 0, 1, 2 ]
    _y_ = [ 0, 1, 2, 2, 0, 0, 1, 2, 1 ]
    # values
    _z_ = [ 1, 1, 1, 1, 1, 0, 0, 0, 0 ]

    (
    uplt.figure(engine.value)
        .surface3d(_x_, _y_, _z_, name='data #1', colormap='hsv', opacity=0.5, show_colormap=True)
        .surface3d(_x_, _y_, _z_, name='data #2',  interpolation='linear', show_colormap=True)
        .xlabel('X Axis')
        .ylabel('Y Axis')
        .zlabel('Z Axis')
        .legend()
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # bar
    """)
    return


@app.cell
def _(engine, np, uplt):
    _y1_ = np.arange(5) + 1


    uplt.figure(engine.value).bar(_y1_, color='r').show()
    return


@app.cell
def _(engine, np, uplt):
    _y1_ = np.arange(5) + 1
    _y2_ = _y1_[::-1]
    _x_ = [ 'A', 'B', 'C', 'AABB', 'D' ]


    (
    uplt.figure(engine.value)
        .bar(_x_, _y1_, name='T1')
        .bar(_x_, _y2_, name='T2')
        .bar(_x_, [1, 1, 1, 1, 1], opacity=0.5, name='T3')
        .ylim(0, 8)
        .hline(3, color='r', line_style='--')
        .legend()
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # hline & vline
    """)
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.linspace(0, 8*np.pi, num=200)
    _y_ = np.cos(_x_)

    (
    uplt.figure(engine.value)
        .plot(_x_, _y_, name='cos')

        .ylim(-2, 2)
        .xlim(-2*np.pi)

        .hline(y=0.5, x_min=10, x_max=15, line_style='--')
        .hline(y=-0.3, line_style='--')

        .vline(x=np.pi+np.pi/2, y_min=-0.5, y_max=0.5, line_style='--')
        .vline(x=4*np.pi, line_style='--')

        .xlabel('X Axis')
        .ylabel('Y Axis')

        .legend()
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # log-scale
    """)
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.linspace(0, 100000, num=1000)
    _y_ = np.sqrt(_x_)


    (
    uplt.figure(engine.value)
        .plot(_x_, _y_, name='sqrt(x)')
        .xscale('log', base=10)
        .yscale('log', base=10)
        .hline(y=100, line_style='--', color='r')
        .vline(x=10000, line_style='--', color='r')
        .xlim(100)
        .ylim(10)
        .legend()
        .show()
    )
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.linspace(0, 100000, num=1000)
    _y_ = np.sqrt(_x_)


    (
    uplt.figure(engine.value)
        .plot(_x_, _y_, name='sqrt(x)')
        .xscale('log', base=2)
        .xlim(2)
        .hline(y=150, line_style='--', color='r')
        .legend()
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # legend
    """)
    return


@app.cell
def _(engine, np, uplt):
    _x_ = np.arange(8)
    _y_ = np.arange(8)


    (
    uplt.figure(engine.value)
        .scatter(_x_, _y_, marker_style='.', marker_size=1, name='data #1', legend_group='Small')
        .scatter(_x_, _y_+1, marker_style='s', marker_size=6, name='data #2', legend_group='Small')
        .scatter(_x_, _y_-1, marker_style='v', marker_size=10, name='data #3', legend_group='Small')
        .scatter(_x_, _y_+3, marker_style='x', marker_size=25, name='data #4', legend_group='Big')
        .plot(_x_, _y_-3, marker_style='o', marker_size=25, name='data #5', legend_group='Big')
        .legend(True, equal_marker_size=True)
        .show()
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # heatmap
    """)
    return


@app.cell
def _(engine, mo, np, uplt):
    # Generate heatmap for visualization
    _range_ = np.linspace(-10, 10, num=101, endpoint=True)
    _xx_, _yy_ = np.meshgrid(_range_, _range_, indexing='xy')
    _z_ = _xx_**3 + _yy_**3


    mo.vstack([
        mo.hstack([ mo.md('## Vertical Colorbar'), mo.md('## Horizontal Colorbar') ],  justify='space-around'),

        mo.hstack([
            uplt.figure(engine.value, width=400, aspect_ratio=0.8).heatmap(_z_, colorbar='vertical').show(),
            uplt.figure(engine.value, width=500, aspect_ratio=0.8).heatmap(_z_, colorbar='horizontal').show(),
        ])
    ])
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # imshow
    """)
    return


@app.cell(hide_code=True)
def _():
    from PIL import Image
    return (Image,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Color Image
    """)
    return


@app.cell
def _(Image, Path, engine, np, uplt):
    _image_ = Image.open(Path(__file__).parent.parent / 'logo.png')
    _image_ = np.array(_image_)


    uplt.figure(engine.value, width=250).imshow(_image_).show()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Grayscale Image
    """)
    return


@app.cell
def _(Image, Path, engine, uplt):
    _image_ = Image.open(Path(__file__).parent.parent / 'logo.png').convert('L')


    uplt.figure(engine.value, width=250).imshow(_image_).show()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # plugin
    """)
    return


@app.cell
def _(np):
    import pandas as pd
    import uplt.plugin as plugin


    class DataFramePlugin(plugin.IPlotPlugin):
        """
        DataFrame minimalistic plugin (only data extraction)
        """

        def extract_data(self, obj: pd.DataFrame) -> list[plugin.PlotData]:
            data = []
            for name in obj.columns:
                if not np.issubdtype(obj.dtypes[name], np.number):
                    continue
                y = np.asarray(obj[name].values)
                x = np.arange(len(y))
                name = name.replace('_', ' ').title()
                data.append(plugin.PlotData(x=x, y=y, name=name))
            return data


    # register plugin
    plugin.register(pd.DataFrame, handler=DataFramePlugin())
    return (pd,)


@app.cell
def _(engine, pd, uplt):
    _car_crashes_ = pd.read_csv(
        'https://raw.githubusercontent.com/mwaskom/seaborn-data/master/car_crashes.csv'
    )

    (
    uplt.figure(engine.value)
        .plot(_car_crashes_[['total', 'speeding', 'alcohol', 'no_previous']])
        .legend().show()
    )
    return


if __name__ == "__main__":
    app.run()



from typing import Literal


ColormapMode = Literal[
    'off',
    'vertical',
    'horizontal'
]


AxisScale = Literal[
    'linear',
    'log',
]


AspectMode = Literal[
    'auto',
    'equal' # equal scale for all axis
]


LineStyle = Literal[
     '-', # solid line
    '--', # dashed line
     ':', # dotted line
    '-.', # dash-dot line
     ' ', # no line
]

MarkerStyle = Literal[
    '.', # point
    ',', # pixel
    'o', # circle
    'v', # triangle down
    '^', # triangle up
    '<', # triangle left
    '>', # triangle right
    '1', # tri down
    '2', # tri up
    '3', # tri left
    '4', # tri right
    '8', # octagon
    's', # square
    'p', # pentagon
    '*', # star
    'h', # hexagon1
    'H', # hexagon2
    '+', # plus
    'x', # x
    'X', # x filled
    'D', # diamond
    'd', # thin diamond
    '|', # vline
    '_', # hline
    'P', # plus filled
]

"""
Common colormaps:
- plotly: https://plotly.com/python/builtin-colorscales/
- matplotlib: https://matplotlib.org/stable/users/explain/colors/colormaps.html
"""
Colormap = Literal[
    'magma',
    'inferno',
    'plasma',
    'viridis',
    'cividis',
    'twilight',
    'turbo',
    'Blues',
    'BrBG',
    'BuGn',
    'BuPu',
    'GnBu',
    'Greens',
    'Greys',
    'OrRd',
    'Oranges',
    'PRGn',
    'PiYG',
    'PuBu',
    'PuBuGn',
    'PuOr',
    'PuRd',
    'Purples',
    'RdBu',
    'RdGy',
    'RdPu',
    'RdYlBu',
    'RdYlGn',
    'Reds',
    'Spectral',
    'YlGn',
    'YlGnBu',
    'YlOrBr',
    'YlOrRd',
    'gray',
    'hot',
    'hsv',
    'jet',
    'ocean',
    'rainbow',
]



from __future__ import annotations

from numpy import ndarray
from pathlib import Path
from numpy.typing import ArrayLike
from collections.abc import Sequence
from abc import abstractmethod as abstract
from typing import Literal, Any, Protocol, runtime_checkable

from uplt.utool import Interpolator
from uplt.utype import LineStyle, MarkerStyle, AspectMode, AxisScale
from uplt.utype import Colormap, ColormapMode


@runtime_checkable
class IFigure(Protocol):
    """
    Matplotlib-like interface for plotting.

    Note
    ----
    This interface follows a Matplotlib-like interface for plotting, providing a familiar
    environment for users, particularly those accustomed to MATLAB.
    It may not be the best choice, but it offers common conventions and practices.
    """

    @property
    @abstract
    def engine(self) -> IPlotEngine:
        """
        Get the underlying plotting engine associated with the figure.

        Returns
        -------
        IPlotEngine
            The plotting engine.
        """

    @property
    @abstract
    def internal(self) -> Any | None:
        """
        Access the underlying engine-specific figure object.
        Use it if you need richer functionality than uplot can provide.
        But manipulation with the internal figure may lead to undefined behavior of uplot.

        Returns
        -------
        Any
            An engine-specific figure object or None if figure is closed.
        """

    @property
    @abstract
    def is_3d(self) -> bool:
        """
        Check if the figure is set up for 3D plotting.

        Returns
        -------
        bool
            True if the figure is in 3D mode, False otherwise.
        """

    @abstract
    def plot(self, x           : ArrayLike | Any,
                   y           : ArrayLike | None = None,
                   z           : ArrayLike | None = None,
                   name        : str | None = None,
                   color       : str | Sequence[str] | None = None,
                   line_style  : LineStyle | None = None,
                   marker_style: MarkerStyle | None = None,
                   marker_size : float | None = None,
                   opacity     : float = 1.0,
                   legend_group: str | None = None,
                   **kwargs) -> IFigure:
        """
        Plot 2D or 3D line.

        Parameters
        ----------
        x, y, z : ArrayLike
            1D data arrays of the same size.
            The x could be any object supported by a plugin.

        name : str or None, optional
            The plot name, which will appear as the legend item.

        color : str or Sequence[str] or None, optional
            The color of the line. Multiple colors can be provided
            for the case of custom objects if corresponding plugin supports it.

        line_style : LineStyle or None, optional
            The line style.

        marker_style : MarkerStyle or None, optional
            The marker style.

        marker_size : float or None, optional
            The size of the marker.

        opacity : float, optional
            Sets the opacity of the line(s).

        legend_group : str or None, optional
            Sets the legend group for this plot.
            Plots from the same group will be combined in the legend.

        kwargs : dict
            Other keyword arguments are forwarded to the underlying engine.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def scatter(self, x           : ArrayLike | Any,
                      y           : ArrayLike | None = None,
                      z           : ArrayLike | None = None,
                      name        : str | None = None,
                      color       : str | Sequence[str] | None = None,
                      marker_style: MarkerStyle | None = None,
                      marker_size : float | None = None,
                      opacity     : float = 1.0,
                      legend_group: str | None = None,
                      **kwargs) -> IFigure:
        """
        Scatter plot for 2D or 3D data points.

        Parameters
        ----------
        x, y, z : ArrayLike
            1D data arrays of the same size.
            The x could be any object supported by a plugin.

        name : str or None, optional
            The plot name, which will appear as the legend item.

        color : str, list of str, or None, optional
            The color(s) of the markers. Multiple colors can be provided
            for each data point or for corresponding plugin in case of custom object.

        marker_style : MarkerStyle or None, optional
            The marker style.

        marker_size : float or None, optional
            The size of the markers.

        opacity : float, optional
            Sets the opacity of the markers.

        legend_group : str or None, optional
            Sets the legend group for this plot. Plots from the same group will be combined in the legend.

        kwargs : dict
            Other keyword arguments are forwarded to the underlying engine.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def hline(self, y           : float,
                    x_min       : float | None = None,
                    x_max       : float | None = None,
                    name        : str | None = None,
                    color       : str | None = None,
                    line_style  : LineStyle | None = None,
                    opacity     : float = 1.0,
                    legend_group: str | None = None,
                    **kwargs) -> IFigure:
        """
        Plot horizontal line at y.
        It supports 2d figures only.

        Parameters
        ----------
        y : float
            y coordinate where to plot the line.

        x_min, x_max : float or None, optional
            x range limit for the line, if None it will be estimated automatically.

        name : str or None, optional
            The plot name, which will appear as the legend item.

        color : str or None, optional
            The color of the line.

        line_style : LineStyle or None, optional
            The line style.

        opacity : float, optional
            Sets the opacity of the line(s).

        legend_group : str or None, optional
            Sets the legend group for this plot.
            Plots from the same group will be combined in the legend.

        kwargs : dict
            Other keyword arguments are forwarded to the underlying engine.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def vline(self, x           : float,
                    y_min       : float | None = None,
                    y_max       : float | None = None,
                    name        : str | None = None,
                    color       : str | None = None,
                    line_style  : LineStyle | None = None,
                    opacity     : float = 1.0,
                    legend_group: str | None = None,
                    **kwargs) -> IFigure:
        """
        Plot vertical line at x.
        It supports 2d figures only.

        Parameters
        ----------
        x : float
            x coordinate where to plot the line.

        y_min, y_max : float or None, optional
            y range limit for the line, if None it will be estimated automatically.

        name : str or None, optional
            The plot name, which will appear as the legend item.

        color : str or None, optional
            The color of the line.

        line_style : LineStyle or None, optional
            The line style.

        opacity : float, optional
            Sets the opacity of the line(s).

        legend_group : str or None, optional
            Sets the legend group for this plot.
            Plots from the same group will be combined in the legend.

        kwargs : dict
            Other keyword arguments are forwarded to the underlying engine.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def surface3d(self, x            : ArrayLike | Any,
                        y            : ArrayLike | None = None,
                        z            : ArrayLike | None = None,
                        name         : str | None = None,
                        show_colormap: bool = False,
                        colormap     : Colormap = 'viridis',
                        opacity      : float = 1.0,
                        interpolation: Interpolator = 'cubic',
                        interpolation_range: int = 100,
                        legend_group : str | None = None,
                        **kwargs) -> IFigure:
        """
        Plot a surface in 3D space where the color scale corresponds to the z-values.
        Two coordinate formats are supported:
          - Uniform grid: x, y are 1D uniform ranges, z is a 2D array of corresponding values.
          - Non-uniform grid (set of points): x, y, z are 1D arrays of corresponding points' coordinates in 3D space.
            The non-uniform grid will be interpolated to a uniform grid using the specified **interpolator**.

        Parameters
        ----------
        x, y, z : ArrayLike
            Data values. The x could be any object supported by a plugin.

        name : str or None, optional
            The plot name, which will appear as the legend item.

        show_colormap : bool, optional
            Whether the colormap should be visualized as a bar alongside the plot.

        colormap : Colormap, optional
            The colormap for the surface.

        opacity : float, optional
            Sets the opacity of the surface.

        interpolation : Interpolator, optional
            The interpolation method for the case when (x, y, z) is a non-uniform grid.

        interpolation_range : int, optional
            The number of points in the interpolated grid.

        legend_group : str or None, optional
            Sets the legend group for this plot. Plots from the same group will be combined in the legend.

        kwargs : dict
            Other keyword arguments are forwarded to the underlying engine.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def bar(self, x           : ArrayLike,
                  y           : ArrayLike | None = None,
                  name        : str | None = None,
                  color       : str | None = None,
                  opacity     : float = 1.0,
                  legend_group: str | None = None,
                  **kwargs) -> IFigure:
        """
        Create a bar plot using the provided data.

        Parameters
        ----------
        x : ArrayLike
            The x-coordinates of the bars. This can represent categories or numeric values.

        y : ArrayLike or None, optional
            The heights of the bars. If not provided, the values in `x` are assumed to be the heights,
            and the bars will be positioned sequentially on the x-axis.

        name : str or None, optional
            The name of the bar plot, which will appear as the legend item.

        color : str or None, optional
            The color of the bars.

        opacity : float, optional
            Sets the opacity of the bars. Default is 1.0 (fully opaque).

        legend_group : str or None, optional
            Sets the legend group for this plot. Plots from the same group will be combined in the legend.

        kwargs : dict
            Additional keyword arguments forwarded to the underlying plotting engine.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def imshow(self, image: ArrayLike, cmap: Colormap | None = None, **kwargs) -> IFigure:
        """
        Display an image.

        Parameters
        ----------
        image : ArrayLike
            Image data. Supported ranges: double [0, 1], uint8, uint16.

        cmap : Colormap or None, optional
            Colormap for single channel images. The 'gray' colormap is used by default.

        kwargs : dict
            Other keyword arguments are forwarded to the underlying engine.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def heatmap(self, data    : ArrayLike,
                      cmap    : Colormap = 'jet',
                      colorbar: ColormapMode = 'vertical') -> IFigure:
        """
        Display a 2d heatmap.

        Parameters
        ----------
        data : ArrayLike
            2D array of values to display as a heatmap.
            The values range is arbitrary.

        colorbar : {'off', 'vertical', 'horizontal'}, optional
            Colorbar position. Default is 'vertical'.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def title(self, text: str) -> IFigure:
        """
        Set the title of the figure.

        Parameters
        ----------
        text : str
            The title text.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def legend(self, show: bool = True,
                     equal_marker_size: bool = True,
                     **kwargs) -> IFigure:
        """
        Show or hide the legend on the figure.

        Parameters
        ----------
        show : bool, optional
            Whether to show or hide the legend.

        equal_marker_size : bool, optional
            Whether markers in the legend are equal in size.

        kwargs : dict
            Other keyword arguments are forwarded to the underlying engine.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def grid(self, show: bool = True) -> IFigure:
        """
        Show or hide the grid on the figure.

        Parameters
        ----------
        show : bool, optional
            Whether to show or hide the grid.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def xlabel(self, text: str) -> IFigure:
        """
        Set the label for the x-axis.

        Parameters
        ----------
        text : str
            The label text.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def ylabel(self, text: str) -> IFigure:
        """
        Set the label for the y-axis.

        Parameters
        ----------
        text : str
            The label text.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def zlabel(self, text: str) -> IFigure:
        """
        Set the label for the z-axis.

        Parameters
        ----------
        text : str
            The label text.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def xlim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        """
        Set limits for the x-axis.

        Parameters
        ----------
        min_value : float or None, optional
            The minimum value for the x-axis.

        max_value : float or None, optional
            The maximum value for the x-axis.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def ylim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        """
        Set limits for the y-axis.

        Parameters
        ----------
        min_value : float or None, optional
            The minimum value for the y-axis.

        max_value : float or None, optional
            The maximum value for the y-axis.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def zlim(self, min_value: float | None = None,
                   max_value: float | None = None) -> IFigure:
        """
        Set limits for the z-axis.

        Parameters
        ----------
        min_value : float or None, optional
            The minimum value for the z-axis.

        max_value : float or None, optional
            The maximum value for the z-axis.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def xscale(self, scale: AxisScale, base: float = 10) -> IFigure:
        """
        Set the scale for the x-axis.

        Parameters
        ----------
        scale : AxisScale
            The scale for the x-axis: 'linear' or 'log'.

        base : float, optional
            The base for the logarithmic scale, default is 10.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def yscale(self, scale: AxisScale, base: float = 10) -> IFigure:
        """
        Set the scale for the y-axis.

        Parameters
        ----------
        scale : AxisScale
            The scale for the y-axis: 'linear' or 'log'.

        base : float, optional
            The base for the logarithmic scale, default is 10.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def current_color(self) -> str:
        """
        Get the color which will be used for the next plot.

        Returns
        -------
        str
            The current color.
        """

    @abstract
    def scroll_color(self, count: int = 1) -> str:
        """
        Scroll a list of predefined colors for plots.

        Parameters
        ----------
        count : int, optional
            The number of colors to scroll.

        Returns
        -------
        str
            The current color before scrolling.
        """

    @abstract
    def reset_color(self) -> IFigure:
        """
        Set the current color to the start of the list.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def axis_aspect(self, mode: AspectMode) -> IFigure:
        """
        Set the aspect ratio of the axis.

        Parameters
        ----------
        mode : AspectMode
            The aspect ratio mode.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def as_image(self) -> ndarray:
        """
        Get the figure as a numpy array.

        Returns
        -------
        ndarray
            The figure as an image.
        """

    @abstract
    def save(self, filename: str | Path) -> IFigure:
        """
        Save the figure to a file.

        Parameters
        ----------
        filename
            The filename for saving the figure.

        Returns
        -------
        IFigure
            The figure object representing the plot.
        """

    @abstract
    def close(self):
        """
        Close the figure. Free allocated resources.
        """

    @abstract
    def show(self, block: bool = True):
        """
        Display the figure.

        Parameters
        ----------
        block : bool, optional
            Whether to block further execution until the figure window is closed.
        """


@runtime_checkable
class IPlotEngine(Protocol):
    """
    Interface for a plotting engine (lib) that provides functionality for creating IFigure.
    """

    @property
    @abstract
    def name(self) -> str:
        """
        Get the name of the plotting engine.

        Returns
        -------
        str
            The name of the plotting engine.
        """

    @classmethod
    @abstract
    def is_available(cls) -> bool:
        """
        Check if the plotting engine is available for use (a plotting lib installed).

        Returns
        -------
        bool
            True if the plotting engine is available, False otherwise.
        """

    @abstract
    def figure(self, width: int, aspect_ratio: float) -> IFigure:
        """
        Factory method for creating a new figure with the specified width and aspect ratio.

        Parameters
        ----------
        width : int
            The width of the figure in pixels.

        aspect_ratio : float
            The aspect ratio of the figure.

        Returns
        -------
        IFigure
            A new figure instance.
        """



from typing import Sequence
from typing import TypeVar


# generic type
T = TypeVar('T')

def unpack_param(param: T | list[T] | None, idx: int) -> T | None:
    """
    Get param by index:
        * param is None -> None
        * param is T -> param
        * param is list[T] -> param[idx]
        * param is list[T] -> None if out of range
    """
    # default
    if param is None:
        return None

    # one param specified
    if isinstance(param, str):
        return param

    # multiple parameters, return one by index
    assert isinstance(param, Sequence)
    if idx < len(param):
        return param[idx]

    # few items parametrized but not all
    return None


from uplt.plugin.IPlotPlugin import IPlotPlugin, PlotData, PlotType
from uplt.plugin.IPlotPlugin import plot
from uplt.plugin.manage import register, is_registered, get_handler


__all__ = [

    # interfaces

    'IPlotPlugin',
    'PlotData',
    'PlotType',

    # functions

    'plot',
    'register',
    'is_registered',
    'get_handler'
]



import importlib.util
from typing_extensions import override # Python 3.12 feature
from uplt.interface import IPlotEngine, IFigure


class MatplotEngine(IPlotEngine):
    # automatically (default) chosen matplotlib backend
    AUTOMATIC_MPL_BACKEND: str | None = None

    @property
    @override
    def name(self) -> str:
        return f'matplotlib-{self._backend.title().lower()}'

    @classmethod
    @override
    def is_available(cls) -> bool:
        return importlib.util.find_spec("matplotlib") is not None

    @property
    def plt(self):
        return self._plt

    @property
    def mpl(self):
        return self._mpl

    @property
    def is_ipython_backend(self) -> bool:
       return ('inline' in self.mpl.get_backend() or
               'ipympl' in self.mpl.get_backend())

    @property
    def is_gui_backend(self) -> bool:
        return not self.is_ipython_backend and self.mpl.get_backend() != 'agg'


    # noinspection PyPackageRequirements
    def __init__(self, backend: str | None = None):
        import matplotlib as mpl
        from matplotlib import pyplot as plt

        self._mpl = mpl
        self._plt = plt

        if self.AUTOMATIC_MPL_BACKEND is None:
            # save default matplotlib backend for future use
            self.AUTOMATIC_MPL_BACKEND = mpl.get_backend()

        if backend is None:
            backend = self.AUTOMATIC_MPL_BACKEND

        self._backend = backend


    @override
    def figure(self, width: int, aspect_ratio: float) -> IFigure:
        from uplt.engine.MatplotFigure import MatplotFigure

        # use style and backend for our figure only
        # avoid to change global state of matplotlib
        current_backend = self._mpl.get_backend()
        self._mpl.use(backend=self._backend)
        fig = MatplotFigure(self, width=width, aspect_ratio=aspect_ratio) # type: ignore
        self._mpl.use(backend=current_backend)
        return fig



from uplt.utool.param import unpack_param
from uplt.utool.image import image_range, image_encode_base64
from uplt.utool.grid import array_to_grid, Interpolator


__all__ = [

    # functions

    'unpack_param',
    'image_range',
    'image_encode_base64',
    'array_to_grid',

    # types

    'Interpolator'
]



import numpy as np
from typing import overload
from typing import OrderedDict, Sequence


color_long_name = OrderedDict[str, str]([
    ('blue'   , '#1f77b4'),
    ('orange' , '#ff7f0e'),
    ('green'  , '#2ca02c'),
    ('red'    , '#d62728'),
    ('purple' , '#9467bd'),
    ('brown'  , '#8c564b'),
    ('magenta', '#e377c2'),
    ('gray'   , '#7f7f7f'),
    ('yellow' , '#bcbd22'),
    ('cyan'   , '#17becf'),
    ('black'  , '#000000'),
    ('white'  , '#ffffff'),
])

color_short_name = {
    'b': color_long_name['blue'],
    'g': color_long_name['green'],
    'r': color_long_name['red'],
    'c': color_long_name['cyan'],
    'm': color_long_name['magenta'],
    'y': color_long_name['yellow'],
    'k': color_long_name['black'],
    'w': color_long_name['white'],
    'o': color_long_name['orange'],
}


@overload
def name_to_hex(name: None) -> None:
    ...

@overload
def name_to_hex(name: str) -> str:
    ...

def name_to_hex(name: str | None) -> str | None:
    """
    Convert a color name to a hex string.

    Parameters
    ----------
    name : str or None
        The color name.

    Returns
    -------
    str or None
        The hex string corresponding to the color name.
    """
    if name is None:
        return None

    if len(name) == 1:
        if name not in color_short_name:
            raise LookupError(f'{name} is not a valid color name, use: {color_short_name.keys()}')
        return color_short_name[name]

    if name[0] == '#':
        return name

    if name not in color_long_name:
        raise LookupError(f'{name} is not a valid color name, use: {color_long_name.keys()}')

    return color_long_name[name]


def rgb_to_str(rgb: list | np.ndarray) -> str | list[str]:
    """
    Convert an RGB value (or array of RGB values) to a hex-string compatible with the uplot API.

    Parameters
    ----------
    rgb : list or np.ndarray
        RGB value in the [0, 1] range or an array of values (N x 3).

    Returns
    -------
    str or list[str]
        Hex color string or array of strings.
    """
    rgb = np.asarray(rgb)
    assert rgb.shape[-1] == 3, 'Input must be an Nx3 array or list'
    assert rgb.max() <= 1.0, 'RGB value range must be [0, 1]'

    rgb255 = (rgb*255).astype(np.uint8)
    rgb255 = np.atleast_2d(rgb255)
    rgb_str = [ f'#{r:02x}{g:02x}{b:02x}' for r, g, b in rgb255 ]

    if len(rgb) == 1:
        return rgb_str[0]  # single RGB value

    return rgb_str  # RGB array


class ColorScroller:
    """
    Class for maintaining automatic color switching for plotting.
    """
    DEFAULT_COLOR_LIST = (
        name_to_hex('orange'),
        name_to_hex('green'),
        name_to_hex('blue'),
        name_to_hex('red'),
        name_to_hex('purple'),
        name_to_hex('brown'),
        name_to_hex('magenta'),
        name_to_hex('gray'),
        name_to_hex('yellow'),
        name_to_hex('cyan'),
    )


    def __init__(self, color_list: Sequence[str] = DEFAULT_COLOR_LIST):
        """
        Parameters
        ----------
        color_list : list[str]
            List of color hex strings for scrolling.
        """
        self._color_list = color_list
        self._color_index = 0


    def scroll_color(self, count: int = 1) -> str:
        """
        Scroll through the color list.

        Parameters
        ----------
        count : int
            The number of colors to scroll.

        Returns
        -------
        str
            The current color before scrolling.
        """
        current = self.current_color()
        self._color_index += count
        self._color_index %= len(self._color_list)
        return current


    def current_color(self) -> str:
        """
        Return the current color.
        """
        return self._color_list[self._color_index]


    def reset(self):
        """
        Reset the scroller state to the first color in the list.
        """
        self._color_index = 0



from abc import abstractmethod as abstract
from typing import Any, NamedTuple, Sequence, Callable, Literal
from typing import cast, get_args
from types import GenericAlias
from numpy.typing import ArrayLike

import uplt.plugin as plugin


class PlotData(NamedTuple):
    """
    Structure to hold plotting data.
    """
    x: ArrayLike
    y: ArrayLike
    z: ArrayLike | None = None
    name: str | None = None
    group_name: str | None = None


PlotType = Literal['plot', 'scatter', 'surface3d']


class IPlotPlugin:
    """
    Plugin interface to support plotting of custom objects
    """

    @abstract
    def extract_data(self, obj: Any) -> list[PlotData]:
        """
        Extract plotting data (x,y,z) from the object.
        This function must be implemented for minimal object plotting support.
        To add more advanced support (style adjustment), implement `update_style` function.

        Parameters
        ----------
        obj:
            The object for data extraction.

        Returns
        -------
        list[PlotData]
            List of extracted data from the object.
        """
        pass


    def update_style(self, plot_type : PlotType,
                           data_index: int,
                           data_count: int,
                           data_name : str | None,
                           group_name: str | None,
                           **kwargs) -> dict:
        """
        Fine-tunes plotting of the custom object.

        Parameters
        ----------
        plot_type:
            Plotting type.

        data_index:
            Index of the data in list[PlotData], see extract_data.

        data_count:
            Size of list[PlotData], see extract_data.

        data_name:
            Name of the data.

        group_name:
            Name of the data group (interpretation depends on plugin).

        kwargs:
            Current plotting style.
            Style parameters depend on plot_type.

        Returns
        -------
        dict
            Updated style (kwargs).
        """
        if data_name is not None:
            kwargs['name'] = data_name

        return kwargs


def plot(plot_method: Callable,
         x          : ArrayLike | Any,
         y          : ArrayLike | None = None,
         z          : ArrayLike | None = None,
         **kwargs) -> bool:
    """
    Returns True if x is recognized as a custom type with an associated plugin.
    In this case, the data will be automatically extracted from the object and visualized.
    No further actions are needed.
    Otherwise, returns False, and x, y, z are regular arrays.
    """
    # check if x is a custom object or regular arrays
    x_type = get_type(x)
    if y is not None or z is not None:
        return False

    if not plugin.is_registered(x_type):
        return False

    # get registered plugin for x
    handler = plugin.get_handler(x_type)
    assert handler is not None, f'plugin for {x_type} is not registered'

    # extract x,y,z from the object
    data_list = handler.extract_data(x)

    # plot all extracted data
    for i, data in enumerate(data_list):
        plot_method_name = plot_method.__name__
        assert plot_method.__name__ in get_args(PlotType), f'unsupported plot method: {plot_method.__name__}'

        params = handler.update_style(plot_type=cast(PlotType, plot_method_name),
                                      data_index=i,
                                      data_count=len(data_list),
                                      data_name=data.name,
                                      group_name=data.group_name,
                                      **kwargs)
        plot_method(x=data.x, y=data.y, z=data.z, **params)

    return True


def get_type(obj: Any) -> type | GenericAlias:
    """
    Extended version of the type() function.

    This function detects and returns the type of the given object.
    It can also identify GenericAlias for homogeneous arrays, such as list[int] or tuple[float, ...].
    """
    obj_type = type(obj)

    if obj_type is str:
        return str

    if not issubclass(obj_type, Sequence) or len(obj) == 0:
        return obj_type

    item0_type = type(obj[0])
    is_homogeneous = all([type(i) is item0_type for i in obj])

    if not is_homogeneous:
        return obj_type

    if obj_type is list:
        # homogeneous list
        return GenericAlias(list, item0_type)
    elif obj_type is tuple:
        # homogeneous tuple
        return GenericAlias(tuple, (item0_type, ...))
    else:
        raise NotImplementedError(obj_type)



from uplt.engine.MatplotEngine import MatplotEngine
from uplt.engine.PlotlyEngine5 import PlotlyEngine5
from uplt.engine.manage import register, get, available


__all__ = [

    # classes

    'MatplotEngine',
    'PlotlyEngine5',

    # functions

    'register',
    'get',
    'available'
]


# register available engines


if MatplotEngine.is_available():
    # matplotlib GUI
    mpl_gui = MatplotEngine()
    register(engine=mpl_gui, name='matplotlib')
    register(engine=mpl_gui, name='matplot')
    register(engine=mpl_gui, name='mpl')

    # matplotlib without GUI (save to file only)
    mpl_no_gui = MatplotEngine(backend='agg')
    register(mpl_no_gui, name='matplotlib-nogui')
    register(mpl_no_gui, name='matplot-nogui')
    register(mpl_no_gui, name='mpl-nogui')
    register(mpl_no_gui, name='mpl-ng')
    register(mpl_no_gui, name='mpl-io')
    register(mpl_no_gui, name='mpl-file')

if PlotlyEngine5.is_available():
    plotly5 = PlotlyEngine5()
    register(engine=plotly5, name='plotly')
    register(engine=plotly5, name='plotly5')
    register(engine=plotly5, name='pl')
    register(engine=plotly5, name='pl5')



import numpy as np
from numpy.typing import ArrayLike
from collections.abc import Sequence

import uplt.color as ucolor
from uplt.default import DEFAULT
from uplt.interface import LineStyle, MarkerStyle


def plot_line_marker(axis,
                     color       : str | Sequence[str],
                     x           : ArrayLike,
                     y           : ArrayLike | None = None,
                     z           : ArrayLike | None = None,
                     name        : str | None = None,
                     line_style  : LineStyle | list[LineStyle] | None = None,
                     marker_style: MarkerStyle | list[MarkerStyle] | None = None,
                     marker_size : float | None = None,
                     opacity     : float = 1.0,
                     **kwargs):
    """
    General plot: line, line+markers, markers(scatter).
    """
    x = np.atleast_1d(np.asarray(x))

    if y is None:
        y = x
        x = np.arange(len(y))
    else:
        y = np.asarray(y)

    y = np.atleast_1d(y)

    assert x.ndim == y.ndim == 1, 'the input must be 1d arrays'
    assert len(x) == len(y), 'the length of the input arrays must be the same'

    if z is not None:
        z = np.atleast_1d(np.asarray(z))
        assert z.ndim == 1, 'the input must be 1d arrays'
        assert len(x) == len(z), 'the length of the input arrays must be the same'
        plot_data = x, y, z
    else:
        plot_data = x, y

    if marker_size is None:
        marker_size = DEFAULT.marker_size

    if isinstance(color, str):
        color = ucolor.name_to_hex(color)
    else:
        # color specified for each point (x, y)
        color = [ ucolor.name_to_hex(c) for c in color ]

    if line_style == ' ':  # only markers (scatter mode)
        axis.scatter(*plot_data,
                     color=color,
                     label=name,
                     marker=marker_style,
                     s=marker_size ** 2,
                     alpha=opacity,
                     **kwargs)
    else:
        axis.plot(*plot_data,
                  color=color,
                  label=name,
                  marker=marker_style,
                  markersize=marker_size,
                  linestyle=line_style,
                  alpha=opacity,
                  **kwargs)



from uplt.interface import LineStyle, MarkerStyle
from typing import Literal


LINE_STYLE_MAPPING: dict[LineStyle | Literal[''] | None, str | None] = {
    '-' : 'solid',
    '--': 'dash',
    ':' : 'dot',
    '-.': 'dashdot',
    ''  : None,
    ' ' : ' ',
    None: None,
}

MARKER_STYLE_MAPPING: dict[MarkerStyle | None, str | None] = {
    '.': 'circle',
    ',': None,
    'o': 'circle-open',
    'v': 'triangle-down',
    '^': 'triangle-up',
    '<': 'triangle-left',
    '>': 'triangle-right',
    '1': 'y-down',
    '2': 'y-up',
    '3': 'y-left',
    '4': 'y-right',
    '8': 'octagon',
    's': 'square',
    'p': 'pentagon',
    '*': 'star',
    'h': 'hexagon',
    'H': 'hexagon2',
    '+': 'cross-thin',
    'x': 'x-thin',
    'X': 'x',
    'D': 'diamond',
    'd': 'diamond-tall',
    '|': 'line-ns',
    '_': 'line-ew',
    'P': 'cross',
    None: None,
}



# copied from ggplot2, the layout is modified to match BMH style
bmh = {
    'data': {'bar': [{'error_x': {'color': 'rgb(51,51,51)'},
                      'error_y': {'color': 'rgb(51,51,51)'},
                      'marker': {'line': {'color': 'rgb(237,237,237)', 'width': 0.5},
                                 'pattern': {'fillmode': 'overlay', 'size': 10, 'solidity': 0.2}},
                      'type': 'bar'}],
             'barpolar': [{'marker': {'line': {'color': 'rgb(237,237,237)', 'width': 0.5},
                                      'pattern': {'fillmode': 'overlay', 'size': 10, 'solidity': 0.2}},
                           'type': 'barpolar'}],
             'carpet': [{'aaxis': {'endlinecolor': 'rgb(51,51,51)',
                                   'gridcolor': 'white',
                                   'linecolor': 'white',
                                   'minorgridcolor': 'white',
                                   'startlinecolor': 'rgb(51,51,51)'},
                         'baxis': {'endlinecolor': 'rgb(51,51,51)',
                                   'gridcolor': 'white',
                                   'linecolor': 'white',
                                   'minorgridcolor': 'white',
                                   'startlinecolor': 'rgb(51,51,51)'},
                         'type': 'carpet'}],
             'choropleth': [{'colorbar': {'outlinewidth': 0,
                                          'tickcolor': 'rgb(237,237,237)',
                                          'ticklen': 6,
                                          'ticks': 'inside'},
                             'type': 'choropleth'}],
             'contour': [{'colorbar': {'outlinewidth': 0,
                                       'tickcolor': 'rgb(237,237,237)',
                                       'ticklen': 6,
                                       'ticks': 'inside'},
                          'colorscale': [[0, 'rgb(20,44,66)'], [1,
                                         'rgb(90,179,244)']],
                          'type': 'contour'}],
             'contourcarpet': [{'colorbar': {'outlinewidth': 0,
                                             'tickcolor': 'rgb(237,237,237)',
                                             'ticklen': 6,
                                             'ticks': 'inside'},
                                'type': 'contourcarpet'}],
             'heatmap': [{'colorbar': {'outlinewidth': 0,
                                       'tickcolor': 'rgb(237,237,237)',
                                       'ticklen': 6,
                                       'ticks': 'inside'},
                          'colorscale': [[0, 'rgb(20,44,66)'], [1,
                                         'rgb(90,179,244)']],
                          'type': 'heatmap'}],
             'histogram': [{'marker': {'pattern': {'fillmode': 'overlay', 'size': 10, 'solidity': 0.2}},
                            'type': 'histogram'}],
             'histogram2d': [{'colorbar': {'outlinewidth': 0,
                                           'tickcolor': 'rgb(237,237,237)',
                                           'ticklen': 6,
                                           'ticks': 'inside'},
                              'colorscale': [[0, 'rgb(20,44,66)'], [1,
                                             'rgb(90,179,244)']],
                              'type': 'histogram2d'}],
             'histogram2dcontour': [{'colorbar': {'outlinewidth': 0,
                                                  'tickcolor': 'rgb(237,237,237)',
                                                  'ticklen': 6,
                                                  'ticks': 'inside'},
                                     'colorscale': [[0, 'rgb(20,44,66)'], [1,
                                                    'rgb(90,179,244)']],
                                     'type': 'histogram2dcontour'}],
             'mesh3d': [{'colorbar': {'outlinewidth': 0, 'tickcolor': 'rgb(237,237,237)', 'ticklen': 6, 'ticks': 'inside'},
                         'type': 'mesh3d'}],
             'parcoords': [{'line': {'colorbar': {'outlinewidth': 0,
                                                  'tickcolor': 'rgb(237,237,237)',
                                                  'ticklen': 6,
                                                  'ticks': 'inside'}},
                            'type': 'parcoords'}],
             'pie': [{'automargin': True, 'type': 'pie'}],
             'scatter': [{'fillpattern': {'fillmode': 'overlay', 'size': 10, 'solidity': 0.2}, 'type': 'scatter'}],
             'scatter3d': [{'line': {'colorbar': {'outlinewidth': 0,
                                                  'tickcolor': 'rgb(237,237,237)',
                                                  'ticklen': 6,
                                                  'ticks': 'inside'}},
                            'marker': {'colorbar': {'outlinewidth': 0,
                                                    'tickcolor': 'rgb(237,237,237)',
                                                    'ticklen': 6,
                                                    'ticks': 'inside'}},
                            'type': 'scatter3d'}],
             'scattercarpet': [{'marker': {'colorbar': {'outlinewidth': 0,
                                                        'tickcolor': 'rgb(237,237,237)',
                                                        'ticklen': 6,
                                                        'ticks': 'inside'}},
                                'type': 'scattercarpet'}],
             'scattergeo': [{'marker': {'colorbar': {'outlinewidth': 0,
                                                     'tickcolor': 'rgb(237,237,237)',
                                                     'ticklen': 6,
                                                     'ticks': 'inside'}},
                             'type': 'scattergeo'}],
             'scattergl': [{'marker': {'colorbar': {'outlinewidth': 0,
                                                    'tickcolor': 'rgb(237,237,237)',
                                                    'ticklen': 6,
                                                    'ticks': 'inside'}},
                            'type': 'scattergl'}],
             'scattermapbox': [{'marker': {'colorbar': {'outlinewidth': 0,
                                                        'tickcolor': 'rgb(237,237,237)',
                                                        'ticklen': 6,
                                                        'ticks': 'inside'}},
                                'type': 'scattermapbox'}],
             'scatterpolar': [{'marker': {'colorbar': {'outlinewidth': 0,
                                                       'tickcolor': 'rgb(237,237,237)',
                                                       'ticklen': 6,
                                                       'ticks': 'inside'}},
                               'type': 'scatterpolar'}],
             'scatterpolargl': [{'marker': {'colorbar': {'outlinewidth': 0,
                                                         'tickcolor': 'rgb(237,237,237)',
                                                         'ticklen': 6,
                                                         'ticks': 'inside'}},
                                 'type': 'scatterpolargl'}],
             'scatterternary': [{'marker': {'colorbar': {'outlinewidth': 0,
                                                         'tickcolor': 'rgb(237,237,237)',
                                                         'ticklen': 6,
                                                         'ticks': 'inside'}},
                                 'type': 'scatterternary'}],
             'surface': [{'colorbar': {'outlinewidth': 0,
                                       'tickcolor': 'rgb(237,237,237)',
                                       'ticklen': 6,
                                       'ticks': 'inside'},
                          'colorscale': [[0, 'rgb(20,44,66)'], [1,
                                         'rgb(90,179,244)']],
                          'type': 'surface'}],
             'table': [{'cells': {'fill': {'color': 'rgb(237,237,237)'}, 'line': {'color': 'white'}},
                        'header': {'fill': {'color': 'rgb(217,217,217)'}, 'line': {'color': 'white'}},
                        'type': 'table'}]},
    'layout': {'annotationdefaults': {'arrowhead': 0, 'arrowwidth': 1},
               'autotypenumbers': 'strict',
               'coloraxis': {'colorbar': {'outlinewidth': 0,
                                          'tickcolor': 'rgb(237,237,237)',
                                          'ticklen': 6,
                                          'ticks': 'inside'}},
               'colorscale': {'sequential': [[0, 'rgb(20,44,66)'], [1, 'rgb(90,179,244)']],
                              'sequentialminus': [[0, 'rgb(20,44,66)'], [1, 'rgb(90,179,244)']]},
               'colorway': ['#F8766D', '#A3A500', '#00BF7D', '#00B0F6', '#E76BF3'],
               'font': {'color': 'rgb(51,51,51)', 'size': 14},
               'title_x': 0.08,
               'margin': { 'l': 10, 'r': 10, 't': 40, 'b': 20 },
               'geo': {'bgcolor': 'white',
                       'lakecolor': 'white',
                       'landcolor': 'rgb(237,237,237)',
                       'showlakes': True,
                       'showland': True,
                       'subunitcolor': 'white'},
               'hoverlabel': {'align': 'left'},
               'hovermode': 'closest',
               'paper_bgcolor': 'white',
               'plot_bgcolor': 'rgb(237,237,237)',
               'polar': {'angularaxis': {'gridcolor': 'white',
                                         'linecolor': 'white',
                                         'showgrid': True,
                                         'tickcolor': 'rgb(51,51,51)',
                                         'ticks': 'outside'},
                         'bgcolor': 'rgb(237,237,237)',
                         'radialaxis': {'gridcolor': 'white',
                                        'linecolor': 'white',
                                        'showgrid': True,
                                        'tickcolor': 'rgb(51,51,51)',
                                        'ticks': 'outside'}},
               'scene': {'dragmode': 'turntable',
                         'xaxis': {'backgroundcolor': 'rgb(237,237,237)',
                                   'gridcolor': 'white',
                                   'gridwidth': 2,
                                   'linecolor': 'white',
                                   'showbackground': True,
                                   'showgrid': True,
                                   'tickcolor': 'rgb(51,51,51)',
                                   'ticks': 'outside',
                                   'zerolinecolor': 'white'},
                         'yaxis': {'backgroundcolor': 'rgb(237,237,237)',
                                   'gridcolor': 'white',
                                   'gridwidth': 2,
                                   'linecolor': 'white',
                                   'showbackground': True,
                                   'showgrid': True,
                                   'tickcolor': 'rgb(51,51,51)',
                                   'ticks': 'outside',
                                   'zerolinecolor': 'white'},
                         'zaxis': {'backgroundcolor': 'rgb(237,237,237)',
                                   'gridcolor': 'white',
                                   'gridwidth': 2,
                                   'linecolor': 'white',
                                   'showbackground': True,
                                   'showgrid': True,
                                   'tickcolor': 'rgb(51,51,51)',
                                   'ticks': 'outside',
                                   'zerolinecolor': 'white'}},
               'shapedefaults': {'fillcolor': 'black', 'line': {'width': 0}, 'opacity': 0.3},
               'ternary': {'aaxis': {'gridcolor': 'white',
                                     'linecolor': 'white',
                                     'showgrid': True,
                                     'tickcolor': 'rgb(51,51,51)',
                                     'ticks': 'outside'},
                           'baxis': {'gridcolor': 'white',
                                     'linecolor': 'white',
                                     'showgrid': True,
                                     'tickcolor': 'rgb(51,51,51)',
                                     'ticks': 'outside'},
                           'bgcolor': 'rgb(237,237,237)',
                           'caxis': {'gridcolor': 'white',
                                     'linecolor': 'white',
                                     'showgrid': True,
                                     'tickcolor': 'rgb(51,51,51)',
                                     'ticks': 'outside'}},
               'xaxis': {'automargin': True,
                         'griddash': 'dot',
                         'gridcolor': '#B2B2B2',
                         'linecolor': '#B2B2B2',
                         'linewidth': 2,
                         'showline': True,
                         'mirror': True,
                         'showgrid': True,
                         'tickcolor': 'rgb(51,51,51)',
                         'ticks': 'inside',
                         'tickwidth': 2,
                         'ticklen': 6,
                         'title': {'standoff': 15},
                         'zeroline': False,
                         'zerolinecolor': 'white' },
               'yaxis': {'automargin': True,
                         'griddash': 'dot',
                         'gridcolor': '#B2B2B2',
                         'linecolor': '#B2B2B2',
                         'linewidth': 2,
                         'showline': True,
                         'mirror': True,
                         'showgrid': True,
                         'tickcolor': 'rgb(51,51,51)',
                         'ticks': 'inside',
                         'tickwidth': 2,
                         'ticklen': 6,
                         'title': {'standoff': 15},
                         'zeroline': False,
                         'zerolinecolor': 'white' }
               }
}



import uplt
import uplt.plugin as plugin

from dataclasses import dataclass


@dataclass
class CustomObj:
    x: list[float]
    y: list[float]
    name: str


class CustomObjPlugin(plugin.IPlotPlugin):

    def extract_data(self, obj: list[CustomObj]) -> list[plugin.PlotData]:
        data = []
        for o in obj:
            data.append(plugin.PlotData(x=o.x, y=o.y, name=o.name))
        return data


data_list = [
    CustomObj(x=[1, 2, 3], y=[5, 6, 7], name='obj1'),
    CustomObj(x=[7, 6, 5], y=[5, 6, 7], name='obj2'),
]

data_tuple = (
    CustomObj(x=[1, 2, 3], y=[5, 6, 7], name='obj1'),
    CustomObj(x=[7, 6, 5], y=[5, 6, 7], name='obj2'),
)

# plugin.register(pd.DataFrame, handler=DataFramePlugin())
plugin.register(list[CustomObj], handler=CustomObjPlugin())
plugin.register(tuple[CustomObj, ...], handler=CustomObjPlugin())

fig = uplt.figure('mpl')
# fig = uplot.figure('plotly')

# regular plot
fig.plot(data_list)
fig.plot(data_tuple)

fig.legend()
fig.show()
